#pragma TextEncoding	= "UTF-8"
#pragma rtGlobals		= 3	
#pragma IgorVersion		= 8.00
#pragma ModuleName		= WaveArthimeticTool
#pragma version			= 5.0
#pragma DefaultTab		= {3,20,4}		// set default tab width in Igor Pro 9 and later

static Constant debugGetTiming = 0
static Constant kVersion = 5.0
static StrConstant kVersionDate = "07/2023"

// --------------------- Project Updater header ----------------------
// If you're using Igor Pro 8 or later and have Project Updater
// installed, this package can check periodically for new releases.
// https://www.wavemetrics.com/project/Updater
static Constant kProjectID = 21539 																// the project node on IgorExchange
static StrConstant ksShortTitle = "Difference Generator"										// the project short title on IgorExchange

//________________________________________________________________________________________________
//	Written by Stephan Thuermer - https://www.wavemetrics.com/user/chozo
//	Shift, scale, offset and compare data against reference data
//
//	Start the user interface via WaveArithmeticUI(wave1,refwave=wave2). Wave 1 will be shifted and scaled,
//	and then compared against wave 2. Wave 2 is optional and can be selected later in the user interface.
//
//	It is possible to compare waves in user code using the function:
//		int error_code = doWaveArithmetic(WaveArithmeticStruct)
//	Here the structure WaveArithmeticStruct contains all the relevant parameters and wave references (see below).
//	Initialize the structure with WaveArithmeticInitialize() and then modify the desired parameters.
//	
// 2021-02-04 - ver. 4.20:	Initial public release.
// 2021-02-05 - ver. 4.30:	Added multiplication and division modes.
// 2021-03-07 - ver. 4.31:	Forgot to remove debug print command.
// 2021-03-17 - ver. 4.32:	Gauss convolution was not working well if data included NaNs.
//							Now the smoothing value is also displayed in the legend.
// 2021-03-19 - ver. 4.33:	Fixed an error message when loading specific old data sets.
// 2021-08-14 - ver. 4.40:	Improved edge behavior of the Gaussian smoothing (data edge are less likely to bulge).
//							Moved all controls into an external panel.
//							Added controls to directly set absolute X,Y, and scale values for changes.
//							Now Background Pinning honors smoothing as well.
// 2022-03-22 - ver. 4.50:	Fixed wrong panel size for windows.
//							Fixed bug: The arrows were reversed for some panel controls with negative delta scaling.
//							Fixed bug: Out-of-range error could happen for reversed (negative delta) waves.
//							Now multiple instances of the tool can be opened at the same time.
//							Added support for keyboard-shortcut (and other future) settings via external text file.
//							Improved backwards compatibility with older panels.
//							Comparisons can now be generated via user code by invoking WaveArithmeticKernel().
//							Axis split between comparison and result spectrum can now be dragged to a new position.
//							Added controls to for selecting and saving additional data generated from the result.
//							Added keyboard shortcuts for smoothing the input (home/end) and reference (alt+home/end) waves.
// 2022-07-14 - ver. 4.55:	Fixed error when loading very small wave.
//							Added percent-difference mode.
//							Minor big-fixes and code improvements.
// 2023-05-08 - ver. 4.56:	Fixed bug: The buttons of the inset panel could appear on unrelated windows in some circumstances.
//							Fixed bug: Some buttons were not sized properly for Mac.
//							Last version with Igor 6 compatibility.
// 2023-07-31 - ver. 5.00:	Implemented 2D wave support for WaveArithmeticKernel().
//							doWaveArithmetic() (new function name for WaveArithmeticKernel()) outputs error codes instead of string messages.
//							Added full 2D wave support to the user interface.
//							Various bug-fixes and tweaks.
//							Upgraded code to Igor 8.
//							Added support for custom arithmetic expressions.
//________________________________________________________________________________________________

static StrConstant kModEnd	= "_mod"															// label for modified data
static StrConstant kAuxEnd	= "_smt;_int;_der;"													// label for additional data
static StrConstant kOutEnd	= "_calc;_dif;_sum;_mul;_div;_pdif;"								// label for output data
static StrConstant kModePresets = "M - R;M + R;M * R;M / R;(M-R)/(M+R) * 200;"					// presets for the different working modes (since version 5.0)

// the following constants define the string keys for saving and loading parameters from wave notes
static StrConstant kFormulaKey	= "Expression"
static StrConstant kScaleKey	= "Mod Scale"
static StrConstant kAxScaleKey	= "Mod Axis Scale"
static StrConstant kShiftKey	= "Mod Shift"
static StrConstant kOffsetKey	= "Mod Offset"
static StrConstant kModSmthKey	= "Mod Smooth"
static StrConstant kRefSmthKey	= "Ref Smooth"
static StrConstant kXCsrSetKey	= "Cursor Set"
static StrConstant kModeInvKey	= "Result inverted"
static StrConstant kPinBaseKey	= "Baseline scaled"
static StrConstant kRefNameKey	= "Reference name used"

//################################################################################################

Structure WaveArithmeticStruct
	variable mode			// fixed mode of comparison: 1: Subtract (M-R) 2: Add (M+R) 3: Multiply (M*R) 4: Divide (M/R) 5: Percentage Diff => (optional; default is 1)
	char formula[100]		// free-form arithmetic expression containing M and R as stand-in for orgWave and refWave; use instead of mode
	Wave orgWave			// [mandatory] wave to compare 1 - will NOT be modified in any way
	Wave refWave			// [mandatory] wave to compare 2 - will NOT be modified in any way
	Wave cmpWave			// [optional] result of comparison between orgWave and refWave will be written here
	Wave modWave			// [optional] modified wave - saves the scaled and shifted version of orgWave
	Wave modRefWave			// [optional] modified reference wave - saves the smoothed version of refWave (if refSmooth > 0)
	// all parameters are optional
	variable orgSmooth		// amount of smoothing for wave 1					(default 0)
	variable refSmooth		// amount of smoothing for wave 2					(default 0)
	variable baselineXval	// x position of baseline							(default NaN)
	variable baselineYval	// y value (1D) or position (2D)of baseline			(default 0)
	variable baselineZval	// for 2D: z value of baseline						(default 0)
	variable baselineFixed	// baseline is omitted from Y scaling if set to 1	(default 0)
	variable invert			// invert the output in Y							(default 0)
	variable xShift			// modify wave 1: x displacement (in wave's units)	(default 0)
	variable yShift			// modify wave 1: y displacement (in wave's units)	(default 0)
	variable valShift		// modify wave 1: intesity displacement (offset)	(default 0)		
	variable xScale			// modify wave 1: x scale							(default 1)
	variable yScale			// modify wave 1: y scale							(default 1)
	variable valScale		// modify wave 1: y / z scaling						(default 1)
	variable UIpinToggle	// options related to user interface; don't use		(default 0)
	variable UIoptions		// bit options related to user interface; don't use	(default 0)
EndStructure

function WaveArithmeticInitialize(STRUCT WaveArithmeticStruct &s)
	s.mode			= 1
	s.formula		= ""
	s.orgSmooth		= 0
	s.refSmooth		= 0
	s.baselineXval	= NaN
	s.baselineYval	= 0
	s.baselineZval	= 0
	s.baselineFixed	= 0
	s.invert		= 0
	s.xShift		= 0
	s.yShift		= 0
	s.xScale		= 1
	s.yScale		= 1
	s.valShift		= 0
	s.valScale		= 1
	s.UIpinToggle	= 0
	s.UIoptions		= 0
end

//################################################################################################

static Constant err_NoError				= 0
static Constant err_SourceWaveMissing	= -1
static Constant err_WrongWaveType		= -2
static Constant err_WaveDimMismatch		= -3
static Constant err_InvalidMode			= -4
static Constant err_HasNoOverlap		= -5
static Constant err_NoComparisonDone	= -6
static Constant err_EquationInvalid		= -7

function doWaveArithmetic(STRUCT WaveArithmeticStruct &s)
	if (debugGetTiming)
		int timerRef = StartMSTimer
	endif
	//++++++++++++++++++++++++++++++ error handling ++++++++++++++++++++++++++++++++++++++++++++++
	if (!WaveExists(s.orgWave))
		return err_SourceWaveMissing
	endif
	if (WaveType(s.orgWave,1) != 1 || (WaveExists(s.refWave) && WaveType(s.refWave,1) != 1) || WaveDims(s.orgWave) > 2 || (WaveExists(s.refWave) && WaveDims(s.refWave) > 2))
		return err_WrongWaveType
	endif
	if ((s.mode < 1 || s.mode > min(ItemsInList(kOutEnd),ItemsInList(kModePresets))-1) || !strlen(s.formula))
		return err_InvalidMode
	endif
	
	string presets = ReplaceString(" ",kModePresets,"")
	string execCmd = StringFromList(s.mode-1,presets)
	if (strlen(s.formula))
		execCmd = ReplaceString(" ",s.formula,"")
		s.mode  = WhichListItem(execCmd, presets)+1
	else
		s.formula = execCmd
	endif
	
	if (!WaveExists(s.modWave))
		Duplicate/Free s.orgWave, s.modWave
	endif
	if (!WaveExists(s.cmpWave))
		string cmpWaveName = NameOfWave(s.orgWave) + StringFromList(s.mode,kOutEnd)
		Duplicate/O s.orgWave, $cmpWaveName;	Wave s.cmpWave= $cmpWaveName
	endif
	
	if (WaveType(s.modWave,1) != 1 || WaveDims(s.modWave) > 2 || WaveType(s.cmpWave,1) != 1 || WaveDims(s.cmpWave) > 2)
		return err_WrongWaveType
	endif
	if (WaveDims(s.orgWave) != WaveDims(s.modWave) || WaveDims(s.orgWave) != WaveDims(s.cmpWave) || WaveExists(s.refwave) && (WaveDims(s.orgWave) != WaveDims(s.refWave)))
		return err_WaveDimMismatch
	endif
	
	//++++++++++++++++++++++++++++++ modify input wave ++++++++++++++++++++++++++++++++++++++++++
	int is2D = WaveDims(s.orgWave) == 2
	variable smthDrift, intOffset = is2D ? s.baselineZval : s.baselineYval
	
	s.xScale = numtype(s.xScale) != 0 ? 1 : (s.xScale == 0 ? 1 : s.xScale)
	s.yScale = numtype(s.yScale) != 0 ? 1 : (s.yScale == 0 ? 1 : s.yScale)
	s.valScale = abs(s.valScale)
	
	SetScale/P x DimOffset(s.orgWave,0)+s.xShift, DimDelta(s.orgWave,0)*s.xScale, WaveUnits(s.orgWave,0), s.modWave	// X-shift (scaling)
	SetScale/P y DimOffset(s.orgWave,1)+s.yShift, DimDelta(s.orgWave,1)*s.yScale, WaveUnits(s.orgWave,1), s.modWave	// Y-shift (scaling)
	
	s.modWave   = (s.orgWave - intOffset*(s.baselineFixed==1))*s.valScale						// value scaling (with or without min. offset)
	smthDrift   = WaveConvolveGaussHelper(s.modWave, s.orgSmooth)								// smoothing (which can change the min wave value => baseline drift)
	s.valShift += s.UIpinToggle * (intOffset*(1 - s.valScale) - smthDrift)						// if the pin-background state was changed: add or subtract the scaled amount to the offset (prevents 'snapping')
	s.modWave  += s.valShift + (intOffset - smthDrift)*(s.baselineFixed==1)						// value shift (after smoothing)

	//+++++++++++++++++++++++ update the wave notes : save parameters ++++++++++++++++++++++++++++
	string noteStr = "\rWave Modification settings:"
	if (is2D)
		noteStr += "\r"+kShiftKey+"=["+num2str(s.xShift)+","+num2str(s.yShift)+"]"
	else
		noteStr += "\r"+kShiftKey+"="+num2str(s.xShift)
	endif
	if (is2D && (s.xScale != 1 || s.yScale != 1))
		noteStr += "\r"+kAxScaleKey+"=["+num2str(s.xScale)+","+num2str(s.yScale)+"]"
	elseif (s.xScale != 1)
		noteStr += "\r"+kAxScaleKey+"="+num2str(s.xScale)
	endif
	noteStr += "\r"+kOffsetKey+"="	+num2str(s.valShift)
	noteStr += "\r"+kScaleKey+"="	+num2str(s.valScale)
	noteStr += "\r"+kModSmthKey+"="	+num2str(s.orgSmooth)
	
	string currNote = note(s.orgWave)
	Note/K s.cmpWave, noteStr
	Note/K s.modWave, currNote + SelectString(strlen(currNote),"","\r\r") + noteStr
	
	s.cmpWave = NaN; CopyScales/P s.modWave, s.cmpWave											// first reset everything
	//++++++++++++++++++++++++++++++ check reference wave +++++++++++++++++++++++++++++++++++++++
	if (!WaveExists(s.refWave))
		return err_NoComparisonDone
	endif
	
	if ((s.UIoptions & 2^0) != 0 && WaveExists(s.modRefWave))									// skip ref. wave convolution if smoothing parameter did not change
		Wave refWave = s.modRefWave
	else
		if (WaveExists(s.modRefWave))
			Duplicate/O s.refWave, s.modRefWave;	Wave refWave = s.modRefWave
		else
			Duplicate/Free s.refWave tempRef;		Wave refWave = tempRef
		endif
		WaveConvolveGaussHelper(refWave, s.refSmooth)
	endif
	
	variable X1, X2, Y1, Y2
	[X1, X2] = WaveArthimeticTool#findOverlap(s,0)
	[Y1, Y2] = WaveArthimeticTool#findOverlap(s,1)
	if (numType(X2*X1) != 0 || (numType(Y2*Y1) != 0 && is2D))
		return err_HasNoOverlap
	endif
	
	//+++++++++++++++++++++++++++++ now execute wave operation +++++++++++++++++++++++++++++++++++
	execCmd = GetWavesDataFolder(s.cmpWave,2)+"="+ReplaceString("M",ReplaceString("R",execCmd,"$$R$$"),"$$M$$")		// replace with unique character sequence
	if (is2D)
		SetScale/I x X1, X2, WaveUnits(s.cmpWave,0), s.cmpWave
		SetScale/I y Y1, Y2, WaveUnits(s.cmpWave,1), s.cmpWave
		execCmd = ReplaceString("$$M$$",execCmd,"Interp2d("+NameOfWave(s.modWave)+",x,y)")
		execCmd = ReplaceString("$$R$$",execCmd,"Interp2d("+NameOfWave(refWave)  +",x,y)")
	else
		SetScale/I x X1, X2, WaveUnits(s.cmpWave,0), s.cmpWave
		execCmd = ReplaceString("$$M$$",execCmd,GetWavesDataFolder(s.modWave,2)+"(x)")
		execCmd = ReplaceString("$$R$$",execCmd,GetWavesDataFolder(refWave,2)  +"(x)")
	endif
	Execute/Q/Z execCmd
	if (V_flag)
		return err_EquationInvalid
	endif
	s.cmpWave *= s.invert ? -1 : 1

	string waveStr = "Wave Arithmetic\r----------\rCreated from:"
	waveStr += "\rReference Wave="	+ GetWavesDataFolder(s.refWave, 2)
	waveStr += "\rOriginal Wave="	+ GetWavesDataFolder(s.orgWave, 2)
	if (s.mode > 0)
		waveStr += "\r"+StringFromList(s.mode-1,kSetModeKey)
	endif
	waveStr += "\r"+kFormulaKey+"="+s.formula+"\r"
	
	if (numtype(s.baselineXval) == 0)
		if (is2D && (s.baselineYval != 0 || s.baselineZval != 0))
			noteStr += "\r"+kXCsrSetKey+"=["+num2str(s.baselineXval)+","+num2str(s.baselineYval)+"]"
		else
			noteStr += "\r"+kXCsrSetKey+"="+num2str(s.baselineXval)
		endif
	endif
	if (s.refSmooth > 0)
		noteStr += "\r"+kRefSmthKey+"="+num2str(s.refSmooth)
	endif
	if (s.invert)
		noteStr += "\r"+kModeInvKey
	endif
	if (s.baselineFixed == 0)
		noteStr += "\r"+kPinBaseKey
	endif
	
	Note/K s.cmpWave, waveStr + noteStr
	Note/K s.modWave, currNote + SelectString(strlen(currNote),"","\r\r") + waveStr + noteStr

	if (debugGetTiming)	
		Print "Arithmetic calculation took", StopMSTimer(timerRef)/1000, "microseconds"
	endif
	return err_NoError
end

//################################################################################################

static function/S WaveArithmeticErrorMsg(int err)
	Switch (err)
		case err_NoError:
			return "Comparison successful."
		case err_NoComparisonDone:
			return "Source wave was modified, but reference was invalid."
		case err_SourceWaveMissing:
			return "Input wave missing."
		case err_WrongWaveType:
			return "This tool works only for numeric 1D or 2D waves."
		case err_WaveDimMismatch:
			return "Wave dimension mismatch. Choose either only 1D or only 2D waves."
		case err_InvalidMode:
			return "Invalid mode: Set a value between 1 and 5."
		case err_HasNoOverlap:
			return "Input waves have no overlap."
		case err_EquationInvalid:
			return "Could not execute equation. There might be an error with the formula."
		default:
			return "Some unknown error has occurred."
	EndSwitch
end

//################################################################################################

static function [variable p1, variable p2] findOverlap(STRUCT WaveArithmeticStruct &s, int dim)
	variable org, ref
	int mirrorO = sign(DimDelta(s.modWave,dim)) == -1											// mirror for negative delta values => reverses assignment
	int mirrorR = sign(DimDelta(s.refWave,dim)) == -1
	org = mirrorO ? IndexToScale(s.modWave,DimSize(s.modWave,dim)-1,dim) : IndexToScale(s.modWave,0,dim)
	ref = mirrorR ? IndexToScale(s.refWave,DimSize(s.refWave,dim)-1,dim) : IndexToScale(s.refWave,0,dim)
	variable left = org > ref ? org : ref														// biggest value  => left border
	org = mirrorO ? IndexToScale(s.modWave,0,dim) : IndexToScale(s.modWave,DimSize(s.modWave,dim)-1,dim)
	ref = mirrorR ? IndexToScale(s.refWave,0,dim) : IndexToScale(s.refWave,DimSize(s.refWave,dim)-1,dim)
	variable right = org < ref ? org : ref														// smallest value => right border
	if (right < left)
		return [NaN, NaN]																		// no overlap
	endif
	if (mirrorO && mirrorR)
		return [right, left]
	else
		return [left, right]
	endif
end

//############################# convolve with a prepared Gaussian ################################

static function WaveConvolveGaussHelper(Wave in, variable width)
	if (width == 0 || numtype(width) != 0)
		return 0
	endif
	if (debugGetTiming)
		int timerRef = StartMSTimer
	endif	
	variable minVal = WaveMin(in), dX = DimDelta(in,0), dY = DimDelta(in,1)
	int is2D = WaveDims(in)==2,  xDim = DimSize(in,0), yDim = DimSize(in,1)
	variable overhead = is2D ? 5 : 10
	variable add = round(max(abs(overhead*width),11))											// create a wave X times the Gauss width to get the full Gauss in (e.g., 10 = 5 sigma on each side)
	variable wX  = abs(width/DimDelta(in,0))/sqrt(2)											// width normalized by x scaling and convert to expected Gauss 'width'
	variable wY  = abs(width/DimDelta(in,1))/sqrt(2)
	
	int xPnt, yPnt, xAdd, yAdd
	if (is2D)
		xAdd = min(add,2*xDim); yAdd  = min(add,2*yDim)
		xPnt = xDim+2*xAdd;		xPnt += mod(xPnt,2)
		yPnt = yDim+2*yAdd;		yPnt += mod(yPnt,2)												// enlarge the wave by 2*add to get the full convolution range and force even size
		Make/FREE/D/N=(xPnt,yPnt) GaussWave = Gauss(x, xPnt/2, wX, y, yPnt/2, wY)
	else
		add  = min(add,2*xDim+1)																// doesn't have to be larger than two times the full input data
		add += !mod(add,2)																		// force odd size
		Make/FREE/D/N=(add)	   GaussWave = Gauss(x, (add-1)/2, wX)
	endif
	
	if (is2D)																					// create large fill wave
		Make/FREE/D/N=(xPnt,yPnt)	 temp = NaN
//		MultiThread temp[xAdd,xAdd+xDim-1][yAdd,yAdd+yDim-1] = in[p-xAdd][q-yAdd]
		MultiThread temp[xAdd		,xAdd+xDim-1][yAdd		,yAdd+yDim-1] = in[p-xAdd][q-yAdd]
		MultiThread temp[0	 		,xAdd	  -1][yAdd		,yAdd+yDim-1] = in[0     ][q-yAdd]
		MultiThread temp[xAdd+xDim	,xPnt	  -1][yAdd		,yAdd+yDim-1] = in[xDim-1][q-yAdd]
		MultiThread temp[0			,xPnt	  -1][0			,yAdd	  -1] = temp[p][yAdd]
		MultiThread temp[0			,xPnt	  -1][yAdd+yDim	,yPnt	  -1] = temp[p][yAdd+yDim-1]
	else
		Make/FREE/D/N=(xDim+2*add,1) temp = NaN; MultiThread temp[ add, add+xDim-1] = in[p-add]
		MatrixFilter NanZapMedian temp
	endif
	
	variable GaussArea = sum(GaussWave)															// make sure the Gauss area limited even for a few points
	if (GaussArea == 0)
		return 0
	endif
	if (GaussArea > 1)
		GaussWave /= GaussArea
	endif
	if (is2D)
		MatrixOP/Free blurred=IFFT(FFT(temp,2)*FFT(GaussWave,2),3);	ImageTransform swap blurred
		MultiThread in = numtype(in[p][q]) != 0 ? in[p][q]	: blurred[p+xAdd][q+yAdd]
	else
		Convolve/A GaussWave temp
		MultiThread in = numtype(in[p]) != 0 ? in[p] : temp[p+add]
	endif
	
	if (debugGetTiming)
		Print "Convolve calculation of "+NameOfWave(in)+" took", StopMSTimer(timerRef)/1000, "microseconds"
	endif
	return WaveMin(in)-minVal																	// return by how much the baseline has shifted
end

//############################### BACKWARDS COMPATIBILITY ########################################

static Constant	kSaveAux = 0
static Constant	kAuxMode = 1
static StrConstant kSetModeKey	= "Data difference;Data summed;Data multiplied;Data divided;Data percent-difference;"	// save key for all modes. The first one is not used for reloading, since this is the default mode.

function/S WaveArithmeticKernel(STRUCT WaveArithmeticStruct &s)
	if (WaveDims(s.orgWave) == 1)																// structure fields were renamed in ver. 4.6 => transfer to new fields
		if (s.yShift != 0)
			s.valShift = s.yShift; s.yShift = 0
		endif
		if (s.yScale != 0)
			s.valScale = s.yScale; s.yScale = 0
		endif
	endif
	return WaveArithmeticErrorMsg(doWaveArithmetic(s))
end

function SpectraDifferenceGenerator(Wave/Z inwave,[Wave/Z refwave])
	WaveArithmeticUI(inwave,refwave=refwave)
end

//################################################################################################
//												GUI
//################################################################################################

Menu "Spectra Tools"
	WaveArthimeticTool#MenuEntry(),/Q, CreateBrowser; WaveArithmeticUI($GetBrowserSelection(0),refwave=$GetBrowserSelection(1))
end

static function/S MenuEntry()
	int fileID
	string read = ""
	Open/Z/R fileID as ParseFilePath(1,FunctionPath(""),":",1,0)+"Spectra Tools settings.dat"	// just keyboard shortcuts for now.
	if (!V_flag)
		do
			FReadLine fileID, read
		while (strlen(read) > 0 && !StringMatch(read, "start difference generator*"))
		read = ReplaceString("\t", ReplaceString(" ", read, ""), "")
		read = StringByKey(ReplaceString(" ", "start difference generator", ""),read,"=","\r")
		Close fileID
	endif
	return "Compare and Modify Data ..."+SelectString(strlen(read),"","/"+read)
end

Menu "GraphMarquee", dynamic
	WaveArthimeticTool#GraphMarquee(),/Q, GetLastUserMenuInfo; WaveArthimeticTool#MarqueeHandler(WinName(0,1), S_value)
end

static function/S GraphMarquee()
	string gName = WinName(0,1)
	if (StringMatch(gName,"DiffGenMainGraph*") && FindListItem("WaveO",ImageNameList(gName,";")+TraceNameList(gName,";",5)) > -1)
		return "-;Set Region of Interest;Clear Region of Interest;"
	endif
	return ""
end

static function MarqueeHandler(string gName, string menuStr)									// sets or releases cursors I and J
	int isImage = FindListItem("WaveO",ImageNameList(gName,";")) > -1
	int isTrace = FindListItem("WaveO",TraceNameList(gName,";",5)) > -1
	if (!StringMatch(gName,"DiffGenMainGraph*") || (!isImage && !isTrace))
		return -1
	endif
	
	string target
	GetMarquee/W=$gName/K bottom, left
	if (StringMatch(menuStr, "Set*"))
		if (!V_Flag)
			Abort "First drag a marquee on the images to define the Region of Interest.\r\rYou can also select 'Set Region of Interest' directly from the marquee's right-click menu."
		endif
		if (isImage)
			target = SelectString(FindListItem("WaveR",ImageNameList(gName,";"))>-1, "WaveO", "WaveR")
			Cursor/I/H=1/L=1/S=2/C=(0,0,65535)/W=$gName I $target V_left,  V_top
			Cursor/I/H=1/L=1/S=2/C=(65535,0,0)/W=$gName J $target V_right, V_bottom
		else
			target = SelectString(FindListItem("WaveR",TraceNameList(gName,";",5))>-1, "WaveO", "WaveR")
			Cursor/H=2/L=1/S=0/C=(0,0,65535)/W=$gName I $target V_left
			Cursor/H=2/L=1/S=0/C=(65535,0,0)/W=$gName J $target V_right
		endif
	else
		if (strlen(CsrInfo(I,gName)) > 0)
			Cursor/K/W=$gName I
		endif
		if (strlen(CsrInfo(J,gName)) > 0)
			Cursor/K/W=$gName J
		endif
	endif
	return 0
end

//################################################################################################

static StrConstant kModeListKey	= "Subtract (A-B);Add (A+B);Multiply (A*B);Divide (A/B);% Difference;"	// names for modes as appearing in the mode selector of the panel
static StrConstant kAuxListKey	= "None;Smoothed;Integrated;Derivative;"								// names for auxiliary data in the panel 
static StrConstant kWorkingDir	= "Difference_Generator"												// name for the working folder inside Packages

function WaveArithmeticUI(Wave/Z inwave,[Wave/Z refwave])
	if (!WaveExists(inwave))
		return err_SourceWaveMissing
	endif
	if (WaveType(inwave,1) != 1 || (WaveExists(refwave) && WaveType(refwave,1) != 1) || WaveDims(inwave) > 2 || (WaveExists(refwave) && WaveDims(refwave) > 2))
		Abort WaveArithmeticErrorMsg(err_WrongWaveType)
	endif
	if (WaveExists(refwave) && (WaveDims(inwave) != WaveDims(refwave)))
		Abort WaveArithmeticErrorMsg(err_WaveDimMismatch)
	endif
	
	int is2D = WaveDims(inwave) == 2
	//+++++++++++++++++++++++++++++++++ generate working variables ++++++++++++++++++++++++++++++
	string gTitle = "Spectra Arithmetic Tool (ver. "+num2str(kVersion)+" - "+kVersionDate+"): Modifying "
	string gName = UniqueName("DiffGenMainGraph", 6, 1), pName = "DiffGenPanel"
	string pPath = gName+"#"+pName
	string workF = "DF"+ReplaceString("DiffGenMainGraph",gName,"")+"_"+CleanupName(NameOfWave(inwave),0)[0,25]
	NewDataFolder/O root:Packages
	NewDataFolder/O root:Packages:$(kWorkingDir)
	NewDataFolder/O root:Packages:$(kWorkingDir):$(workF)
	DFREF stg = root:Packages:$(kWorkingDir):$(workF)
	
	String/G stg:locOriginal  = GetWavesDataFolder(inwave,2)
	String/G stg:locReference = ""
	Make/D/O stg:Modifiers	  = {0,0,1,0}														// parameters: [0] = horizontal shift, [1] = vertical shift, [2] = multiplier, [3] = z offset (2D data)
	//+++++++++++++++++++++++++++++++++++++ build GUI window +++++++++++++++++++++++++++++++++++++
	variable o = 10, h = 27, i = 0, grp = 0														// panel: offset, height, increment row, group start row
	variable pWidth = 250, pHeight = 50+h*(16+is2D*9)
	variable WinTop = 70,  WinLeft = 250+(200*is2D)

	variable TFTResX, TFTResY, pix = 72/ScreenResolution										// get the display resolution to align window better (center of full screen)
	if (cmpstr(IgorInfo(2),"Macintosh")==0)		// Mac
		string TFTInfo = StringByKey("RECT",ReplaceString(",RECT",StringByKey("SCREEN1",IgorInfo(0),":"),";RECT"),"=")
		TFTResX = (Str2Num(StringFromList(2,TFTInfo,",")) - Str2Num(StringFromList(0,TFTInfo,",")))*(72/ScreenResolution)  
		TFTResY = (Str2Num(StringFromList(3,TFTInfo,",")) - Str2Num(StringFromList(1,TFTInfo,",")))*(72/ScreenResolution)  
	else										// Windows
		GetWindow kwFrameInner wsizeDC
		TFTResX = (v_right-v_left)
		TFTResY = (v_bottom-v_top)
	endif
	variable WinBottom	= TFTResY - 70;			WinBottom = limit(WinBottom, WinTop+300, inf)
	variable WinRight	= TFTResX - 50-pWidth;	WinRight  = limit(WinRight, WinLeft+300, inf)
	
	Display/K=2/W=(WinLeft*pix, WinTop*pix, WinRight*pix, WinBottom*pix)/N=$gName as gTitle + NameOfWave(inwave)
	NewPanel/HOST=$gName/EXT=0/K=2/N=$pName/W=(0, 0, pWidth, pHeight) as "Spectra Modification"
	SetWindow $gName hook(DiffWindow)=DiffGen_ProcessWindowEvent								// hook function processes key events
	SetWindow $pPath hook(DiffWindow)=DiffGen_ProcessWindowEvent								// keeps track of button presses here as well
	SetWindow $pPath userdata(workFolder)=workF													// save work folder location
	SetWindow $gName userdata(procVersion)=num2str(kVersion)									// save procedure version
	
	//++++++++++++++++++++++++++++++ external panel with the controls ++++++++++++++++++++++++++++
	TitleBox FomulaLabel	,pos={30  ,2+o+h*i}	,size={23,23}		,title="Mode:"		,fstyle=1			,frame=0
	PopUpMenu popFormula	,pos={5   ,o+h*i}	,size={20,23}		,title=""							,help={"Pre-selects the operation modes: M-R difference (- key), M+R addition (+ key), M*R multiplication (* key), M/R division (/ key) or (M-R)/(M+R) percentage difference."}
	Setvariable FormulaStr	,pos={70  ,o+h*i}	,size={115,23}		,title=""							,help={"Directly enter arithmetic expression here using both 'M' and 'R' as placeholders for the modified and reference waves, respectively. Example M-R."}
//	PopUpMenu ArithMode		,pos={10  ,o+h*i}	,size={175,23}		,title="Mode:"		,fstyle=1		,help={"Selects the current operation mode: A-B difference (- key), A+B addition (+ key), A*B multiplication (* key), A/B division (/ key) or (A-B)/(A+B) percentage difference."}
	CheckBox InvertMode		,pos={195,1+o+h*i++},size={50 ,23}		,title="\f04I\f00nvert"				,help={"Inverts the difference spectrum. (key: 'i')"}

	Button bProgSwitch		,pos={220 ,o+h*i}	,size={20 ,2*h-9}	,title="]"			,fsize=20		,help={"Switches editing and reference waves."}	
	PopUpMenu OrigSelect	,pos={10  ,o+h*i++}	,size={200,23}		,title="\f01M\f00od.:"				,help={"Select the wave which will be modified."}
	PopUpMenu RefSelect		,pos={10  ,o+h*i++}	,size={200,23}		,title="\f01R\f00ef.:"				,help={"Select the reference wave to compare against."}
	
	grp = i
	GroupBox ConvLabel		,pos={8 ,5+o+h*i++}	,size={235,50}		,title="Smoothing (Gauss convolution):"	,frame=0
	Setvariable WaveConv	,pos={20  ,o+h*i}	,size={105,23}		,title="Mod.:"						,help={"Convolves the modified wave with a Gaussian."}
	Setvariable RefWaveConv	,pos={130 ,o+h*i++}	,size={105,23}		,title="Ref.:"						,help={"Convolves the reference wave with a Gaussian. Modification will not be saved."}
	GroupBox ConvLabel							,size={235,h*(i-grp)-5}

	grp = i
	GroupBox RelChangeLabel	,pos={8 ,5+o+h*i++}	,size={235,100}		,title="Relative changes:"				,frame=0
	Setvariable Shift		,pos={15  ,o+h*i}	,size={130,23}		,title="delta X:"	,bodyWidth=80	,help={"The delta value by which the horizontal position will be shifted every step."}
	Button bUp				,pos={185 ,o+h*i++}	,size={20 ,20}		,title="^"							,help={"Add a constant value to the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: up arrow)"}
	Setvariable Offset		,pos={15  ,o+h*i}	,size={130,23}		,title="delta Y:"	,bodyWidth=80	,help={"The delta value by which the vertical position will be shifted every step."}
	Button bDown			,pos={185 ,o+h*i}	,size={20 ,20}		,title="v"							,help={"Remove a constant value from the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: down arrow)"}
	Button bLeft			,pos={155 ,o+h*i}	,size={20 ,20}		,title="<"							,help={"Shift the spectrum to lower x values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: left arrow)"}
	Button bRight			,pos={213 ,o+h*i++}	,size={20 ,20}		,title=">"							,help={"Shift the spectrum to higher x values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: right arrow)"}
if (is2D)
	Setvariable zOffset		,pos={15  ,o+h*i}	,size={130,23}		,title="delta Z:"	,bodyWidth=80	,help={"The delta value by which the z position will be shifted every step."}
	Button bZUp				,pos={155 ,o+h*i}	,size={35 ,20}		,title="Z ^"						,help={"Add a constant value to the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: up arrow)"}
	Button bZDown			,pos={200 ,o+h*i++}	,size={35 ,20}		,title="Z v"						,help={"Remove a constant value from the data every step. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: down arrow)"}
	Button bUp																							,help={"Shift the spectrum to lower y values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: left arrow)"}
	Button bDown																						,help={"Shift the spectrum to higher y values. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the movement. Very fine movement with both ctrl/command + shift. (key: right arrow)"}
endif
	Setvariable Multiplier	,pos={15  ,o+h*i}	,size={130,23}		,title="Scale by:"					,help={"The intensity will be multiplied by this value every step."}			
	Button bStretch			,pos={155 ,o+h*i++}	,size={80 ,20}		,title="Stretch"					,help={"Multiply intensity by the set percentage. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the stretching. Very fine movement with both ctrl/command + shift. (key: page up)"}
	CheckBox PinBase		,pos={30  ,o+h*i}	,size={80 ,23}		,title="\f04P\f00in down Baseline:"	,help={"The baseline (wave minimum) stays fixed and will not be scaled. The fix point can be selected by setting cursor A or B on the reference wave (WaveR). (key: 'p')"}
	Button bShrink			,pos={155 ,o+h*i++}	,size={80 ,20}		,title="Shrink"						,help={"Downsize intensity by the set percentage. Hold 'shift' ('ctrl/command') to accelerate (decelerate) the shrinking. Very fine movement with both ctrl/command + shift. (key: page down)"}
	GroupBox RelChangeLabel						,size={235,h*(i-grp)-5}
	
	grp = i
	GroupBox AbsChangeLabel	,pos={8 ,5+o+h*i++}	,size={235,100}		,title="Absolute changes:"				,frame=0
	Setvariable AbsXshift	,pos={15  ,o+h*i}	,size={130,23}		,title="X:"							,help={"Set the absolute horizontal shift."}
	Button bUndo			,pos={155 ,o+h*i++}	,size={80 ,20}		,title="\f05R\f01eset:\f00 All"		,help={"Reset all modifications. Hold 'ctrl/command' to additionally unload the reference wave. (key: 'r')"}
	Button bUndoShift		,pos={155 ,o+h*i}	,size={35 ,20}		,title="X"							,help={"Reset horizontal shift."}
	Setvariable AbsYshift	,pos={15  ,o+h*i}	,size={130,23}		,title="Y:"							,help={"Set the absolute vertical shift."}
	Button bUndoOffset		,pos={200 ,o+h*i++}	,size={35 ,20}		,title="Y"							,help={"Reset vertical shift."}
if (is2D)
	Setvariable AbsZshift	,pos={15  ,o+h*i}	,size={130,23}		,title="Z:"							,help={"Set the absolute Z shift."}
	Button bUndoZOffset		,pos={155 ,o+h*i++}	,size={80 ,20}		,title="Z offset"					,help={"Reset Z offset (baseline)."}
endif
	Setvariable AbsScale	,pos={15  ,o+h*i}	,size={130,23}		,title="Scale:"						,help={"Set the absolute scale."}
	Button bUndoScale		,pos={155 ,o+h*i++}	,size={80 ,20}		,title="Scale"						,help={"Reset scaling."}
	GroupBox AbsChangeLabel						,size={235,h*(i-grp)-5}
	
	o += 10
	Button bProgUndoQuit	,pos={10  ,o+h*i}	,size={60 ,2*h-9}	,title="Discard\r& Quit"			,help={"Discard changes and quit."}
	Button bProgSaveQuit	,pos={80  ,o+h*i}	,size={55 ,2*h-9}	,title="\f01Save\r& Quit"			,help={"Save changes and quit. Hold 'ctrl/command' key to just save the result without quitting."}	,valueColor=(65535,0,0)
	CheckBox FinishGraph	,pos={145 ,o+h*i++}	,size={95 ,20}		,title="Display Result"				,help={"Display results graph at the end."}
	CheckBox AltSaveOption	,pos={145 ,o+h*i++}	,size={95 ,20}		,title="Use Ref. Name"				,help={"Saves the difference wave with the name of the reference wave instead of the modified wave."}
	PopUpMenu popAuxMode	,pos={10  ,o+h*i}	,size={125,23}		,title="Show:"	,fstyle=1			,help={"Selects which type of additional data to show."}
	CheckBox cAuxSave		,pos={145 ,o+h*i++}	,size={95 ,20}		,title="Save Aux. Data"				,help={"Also saves the additional data (green curve)."}
	
if (is2D)
	grp = i; o -= 5
	GroupBox ImgCtrlsLabel	,pos={8 ,5+o+h*i++}	,size={235,100}		,title="Image control:"					,frame=0
	TitleBox refAlphaLabel	,pos={20  ,o+h*i+2}	,size={23,23}		,title="Ref. Alpha:"					,frame=0
	Slider ref2D_Alpha		,pos={88  ,o+h*i}	,size={90,23}											,help={"Set the transparency of the reference image."}
	CheckBox ref2D_DoShow	,pos={190 ,o+h*i++}	,size={95 ,20}		,title="\f04S\f00how"				,help={"Toggles the display of the reference image."}
	PopupMenu src2D_Color	,pos={15  ,o+h*i++}	,size={220,20}		,title="Modifying:"					,help={"Select the color of the modified image."}
	PopupMenu ref2D_Color	,pos={15  ,o+h*i++}	,size={220,20}		,title="Reference:"					,help={"Select the color of the reference image."}
	o += 3
	TitleBox RangeLowLabel	,pos={20  ,o+h*i-2}	,size={23,23}		,title="Range Min.:"					,frame=0
	Slider ValRangeLow		,pos={88  ,o+h*i++}	,size={145,10}
	TitleBox RangeHighLabel	,pos={20  ,o+h*i-2}	,size={23,23}		,title="Range Max.:"					,frame=0
	Slider ValRangeHigh		,pos={88  ,o+h*i++}	,size={145,10}
	o -= 3
	Button bProgShowProfile	,pos={18  ,o+h*i}	,size={100,20}		,title="Show Profiles"				,help={"Show horizontal and vertical profiles of the 2D data."}
	Button bProgActivateROI	,pos={130 ,o+h*i++}	,size={105,20}		,title="Set / Clear ROI"			,help={"Sets (or clears) the Region of Interest from an marquee on the graph."}
	GroupBox ImgCtrlsLabel						,size={235,h*(i-grp)-5}
	
	int startColor = 1+WhichListItem("Terrain", CTabList())
	PopupMenu ref2D_Color	,bodyWidth=155	,value="*COLORTABLEPOPNONAMES*"	,mode=startColor	,proc=ImageColorProc
	PopupMenu src2D_Color	,bodyWidth=155	,value="*COLORTABLEPOPNONAMES*"	,mode=startColor	,proc=ImageColorProc
	Slider ValRangeLow		,side=0	,vert=0	,value=0	,ticks=0	,limits={0,1,0}				,proc=RangeSliderProc
	Slider ValRangeHigh		,side=0	,vert=0	,value=1	,ticks=0	,limits={0,1,0}				,proc=RangeSliderProc
	Slider ref2D_Alpha		,side=2	,vert=0	,value=0.5	,ticks=-3	,limits={0,1,0}				,proc=AlphaSliderProc
	CheckBox ref2D_DoShow					,value=1											,proc=ToggleRefVisibility
endif
	
	if (CmpStr(IgorInfo(2), "Macintosh") == 0)
		ModifyControlList ControlNameList(pPath,";","*") fsize=10
	endif
	
	ModifyControlList ControlNameList(pPath, ";", "b*")	,proc=DiffGen_ButtonControl
	
	PopUpMenu OrigSelect	,bodywidth=165	,value=#("getPopWaveList(0,"+num2str(1+is2D)+")")	,proc=DiffGen_PopupWaveSelect	,mode=1	,popvalue=NameOfWave(inwave)
	PopUpMenu RefSelect		,bodywidth=165	,value=#("getPopWaveList(1,"+num2str(1+is2D)+")")	,proc=DiffGen_PopupWaveSelect	,mode=1
	PopUpMenu popFormula	,bodywidth=20	,value=#("\""+kModePresets+"\"")					,proc=DiffGen_ModeSelect		,mode=0
//	PopUpMenu ArithMode		,bodywidth=115	,value=#("\""+kModeListKey+"\"")					,proc=DiffGen_ModeSelect		,mode=1
	PopUpMenu popAuxMode	,bodywidth=85	,value=#("\""+kAuxListKey+"\"")						,proc=DiffGen_ModeSelect		,mode=1+!is2D
	
	Setvariable FormulaStr	,bodyWidth=115	,value=_STR:StringFromList(0,kModePresets)			,proc=DiffGen_WaveConvolve
	Setvariable Multiplier	,bodyWidth=80	,value=_NUM:1				,format="%g %"	,limits={0.01,inf,0.5}
	Setvariable AbsXshift	,bodyWidth=110	,value=stg:Modifiers[0]								,proc=DiffGen_WaveConvolve
	Setvariable AbsYshift	,bodyWidth=110	,value=stg:Modifiers[1]		,format="%g"			,proc=DiffGen_WaveConvolve
	Setvariable AbsScale	,bodyWidth=90	,value=stg:Modifiers[2]								,proc=DiffGen_WaveConvolve
	Setvariable WaveConv	,bodyWidth=75								,format="%g x"			,proc=DiffGen_WaveConvolve
	Setvariable RefWaveConv	,bodyWidth=75	,value=_NUM:0										,proc=DiffGen_WaveConvolve
if (is2D)
	Setvariable AbsZshift	,bodyWidth=110	,value=stg:Modifiers[3]		,format="%g"			,proc=DiffGen_WaveConvolve
endif
	CheckBox InvertMode																			,proc=DiffGen_InvertMode
	CheckBox PinBase		,side=1			,value=1											,proc=DiffGen_PinBaseMode
	CheckBox FinishGraph					,value=1
	//++++++++++++++++++++++ add the traces after the graph was built ++++++++++++++++++++++++++++
	DoWindow/F $gName
	string recreate = LoadWavesAndSettings(pPath)												// create duplicates of the original waves and reload parameters if available (needs to be executed AFTER window controls were built)
	Wave/SDFR=stg WaveO, WaveR, WaveD, WaveA													// WaveO = wave to edit; WaveR = reference wave; WaveD = difference wave; WaveA = auxiliary wave
	
if (is2D)
	AppendImage WaveO			;ModifyImage WaveO ctab= {*,*,Terrain,0}
	AppendImage/L=Ldiff WaveD	;ModifyImage WaveD ctab= {*,*,Terrain,0}
	
	ControlInfo/W=$pPath ref2D_Color															// prepare reference color wave
	ColorTab2Wave $S_value;	Wave M_colors;	Redimension/N=(-1,4) M_colors
	Duplicate/O M_colors, stg:refColors; 	KillWaves/Z M_colors
	ImageSetAlpha(pPath)
else
	AppendToGraph/C=(65535,0,    0) WaveO
	AppendToGraph/C=(0,    0,    0)/L=Ldiff WaveD
	AppendToGraph/C=(0,40000,20000)/L=Ldiff WaveA
endif
	
	variable axDivider = 0.3
	ModifyGraph axisEnab(left)	= {axDivider + 0.01, 1}											// result is displayed at the bottom 30% of the graph
	ModifyGraph axisEnab(Ldiff) = {0, axDivider - 0.01}
	ModifyGraph freePos(Ldiff)=0, zero(Ldiff)=3, mirror=1, minor=1, standoff=0
	ModifyGraph lblPos(Ldiff)=50, ZisZ=1, notation(left)=1, gfSize=10
	SetAxis/A=2 Ldiff;	SetAxis/A=2 left
if (is2D)
	ColorScale/C/N=MainColorDisplay/F=0/Z=1/A=LT/X=100.10/Y=0.00 " "
	ColorScale/C/N=DiffColorDisplay/F=0/Z=1/A=LB/X=100.10/Y=0.00 " "
	ColorScale/C/N=MainColorDisplay image=WaveO
	ColorScale/C/N=DiffColorDisplay image=WaveD
	ColorScale/C/N=MainColorDisplay vert=1, side=1, heightPct=100-(axDivider-0.01)*100,	width=8, tickLen=5
	ColorScale/C/N=DiffColorDisplay vert=1, side=1, heightPct=(axDivider-0.01)*100, 	width=8, tickLen=5
	ColorScale/C/N=MainColorDisplay tickLblRot=90, fsize=10, lblMargin=0, minor=1, tickUnit=1
	ColorScale/C/N=DiffColorDisplay tickLblRot=90, fsize=10, lblMargin=0, minor=1, tickUnit=1
	ModifyGraph margin(right)=40, mirror(left)=2, mirror(Ldiff)=2
endif
	
	Legend/C/N=GraphInfoBox/J/F=0/A=LT/X=2.00/Y=2.00 ""											// trace legend
	Legend/C/N=ModificationInfoBox/J/F=0/A=RT/X=2.00/Y=2.00/B=1 ""								// modification legend
	DoUpdate
	UpdateGraphLegend(pPath)
	
	WaveR = NaN
	if(WaveExists(refwave))																		// if reference wave is provided in function call
		DFREF saveDF = GetDataFolderDFR()
		SetDataFolder GetWavesDataFolderDFR(refwave)
			PopUpMenu RefSelect ,mode=1 ,popvalue=NameOfWave(refwave)
			doWaveSelection(pPath,"RefSelect", NameOfWave(refwave))
		SetDataFolder saveDF
	else																						// instead sets the old reference wave if available
		ReloadRefWaveSettings(pPath, recreate)													// reloads previous reference wave and sets relevant controls
	endif
	
	CalculateResult(pPath)
	
//	if (is2D)
//		ShowProfilePanel(stg, gName)															// add an extra panel with profiles of 2D data (must come last)
//	endif
	return 0
end
//++++++++++++++++++++++++++++++++++++++++ number rounding +++++++++++++++++++++++++++++++++++++++
static function ChangePrecision(variable value, variable numSigDigits)							// round values to given number of digits
	string str; sprintf str, "%.*g\r", numSigDigits, value
	return str2num(str)
end

//################################################################################################

static function/DF getWorkFolder(string pPath)													// grab current working folder from panel info
	string folder = ""
	if (strlen(pPath))
		folder = GetUserData(pPath,"","workFolder")
	endif
	return root:Packages:$(kWorkingDir):$(folder)
end

static function/S getPanel(string gName)
	string subPanel = StringFromList(0,RemoveFromList("AskQuitPanel",ChildWindowList(gName)))	// grab attached panel from list
	return SelectString(strlen(subPanel), "DiffGenGraphMainWindow", gName+"#"+subPanel)			// compatibility with older versions
end

static function/S getGraph(string pPath)
	return StringFromList(0,pPath,"#")
end

//################################################################################################

static function ShowProfilePanel(DFREF stg, string gName)
	string pName = "DiffGen_2DProfiles"
	string pPath = gName+"#"+pName
	if (WinType(pPath) == 7)
		KillWindow $pPath
//		return 0
	endif
	
	GetWindow $gName gsizeDC
	NewPanel/HOST=$gName/EXT=1/K=1/N=$pName/W=(350,0,0,V_bottom-V_top) as "Profile View"		// has the same height as the main graph
	Display/HOST=$pPath/FG=(FL,FT,FR,FB)/N=$("ProfileGraph")
	ModifyPanel/W=$pPath fixedSize=0
	
	Wave/Z/SDFR=stg WaveR_prV, WaveR_prH
	if (!WaveExists(WaveR_prV) || !WaveExists(WaveR_prH))
		Duplicate/O stg:WaveD_prV, stg:WaveR_prV
		Duplicate/O stg:WaveD_prH, stg:WaveR_prH
	endif
	AppendToGraph/C=(65535,0,    0)/L=Lvert/T=Tvert stg:WaveO_prV
	AppendToGraph/C=(0,    0,    0)/R=Rvert/T=Tvert stg:WaveD_prV
	AppendToGraph/C=(0,    0,65280)/L=Lvert/T=Tvert stg:WaveR_prV
	AppendToGraph/C=(65535,0,    0)   stg:WaveO_prH
	AppendToGraph/C=(0,    0,    0)/R stg:WaveD_prH
	AppendToGraph/C=(0,    0,65280)   stg:WaveR_prH

	Label left  "\\f01Input \\U"
	Label Lvert "\\f01Input \\U"
	Label right "\\f01Result \\U"
	Label Rvert "\\f01Result \\U"
	
	SetAxis/A/N=2 right
	SetAxis/A/N=2 Rvert
	NewFreeAxis/B Bdiv																			// these are just divider lines
	NewFreeAxis/T Tdiv
	
	variable axDivider = 0.5, axSpace = 0.01
	ModifyGraph axisEnab(left)	= {0,axDivider-axSpace}, axisEnab(Lvert) = {axDivider+axSpace,1}
	ModifyGraph axisEnab(right)	= {0,axDivider-axSpace}, axisEnab(Rvert) = {axDivider+axSpace,1}
	ModifyGraph freePos(Tdiv)	= {axDivider-axSpace,kwFraction}
	ModifyGraph freePos(Bdiv)	= {axDivider-axSpace,kwFraction}
	ModifyGraph  freePos(Lvert) = 0,  freePos(Rvert)= 0,  freePos(Tvert)= 0
	ModifyGraph tkLblRot(Lvert) =90, tkLblRot(Rvert)=90, tkLblRot(right)=90, tkLblRot(left)=90
	ModifyGraph   lblPos(Lvert) =50,   lblPos(Rvert)=50,   lblPos(Tvert)=50,  lblPos(bottom)=0
	
	ModifyGraph nticks(Bdiv)=0 ,noLabel(Bdiv)=2
	ModifyGraph nticks(Tdiv)=0 ,noLabel(Tdiv)=2
	ModifyGraph mirror(bottom)=2, mirror(Tvert)=2												//, mirror(left)=1, mirror(Lvert)=1
	ModifyGraph minor=1, standoff=0, ZisZ=1, gfSize=12, highTrip=1000,notation=1
	ModifyGraph lstyle(WaveD_prH)=2, lstyle(WaveD_prV)=2
	ModifyGraph zero(Rvert)=4, zero(right)=4
	
	TextBox/C/N=ProfileH/J/F=0/B=1/A=LC/X=5.00/Y=-5.00 "\Zr175\f01X 🡘"
	TextBox/C/N=ProfileV/J/F=0/B=1/A=LT/X=5.00/Y=2.00  "\Zr175\f01Y 🡙\M\f00\r\s(WaveO_prV) Mod.\r\s(WaveR_prV) Ref.\r\s(WaveD_prV) Result"
	
	DoWindow/F $gName
end

//################################################################################################

static function CalculateResult(string pPath, [variable undoPinning])
	if (debugGetTiming)
		int timerRef = StartMSTimer
	endif
	
	STRUCT WaveArithmeticStruct s
	WaveArithmeticInitialize(s)
	
	DFREF stg = getWorkFolder(pPath)
	SVar/SDFR=stg locOriginal, locReference
	Wave/SDFR=stg WaveO, WaveR, WaveD, Modifiers												// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplier, [3] = z offset (2D data)
	Wave/Z s.refWave = $locReference
	Wave   s.orgWave = $locOriginal
	Wave   s.cmpWave = WaveD
	Wave   s.modWave = WaveO
	Wave   s.modRefWave = WaveR
	
	if (DimSize(Modifiers,0) < 4)
		Redimension/N=(4) Modifiers
	endif
	Modifiers = numtype(Modifiers[p]) != 0 ? p == 2 : Modifiers[p]								// first, make sure these are proper numbers
	
	int is2D = WaveDims(WaveO)==2
	ControlInfo/W=$pPath PinBase;		s.baselineFixed = V_Value
	ControlInfo/W=$pPath WaveConv;		s.orgSmooth = V_Value
	ControlInfo/W=$pPath RefWaveConv;	s.refSmooth = V_Value
	ControlInfo/W=$pPath InvertMode;	s.invert = V_Value
	ControlInfo/W=$pPath ArithMode;		s.mode = V_flag==3 ? V_value   : 1						// backwards compatibility => default is 1; obsolete since version 5.0
	ControlInfo/W=$pPath SumMode;		s.mode = V_flag==2 ? V_Value+1 : s.mode					// backwards compatibility => old checkbox
	ControlInfo/W=$pPath FormulaStr;	s.formula = SelectString(V_flag==5,"",S_Value)			// added in version 5.0
	
	variable prevRefSmooth = str2num(StringByKey(kRefSmthKey,note(WaveO),"=","\r"))
			 prevRefSmooth = numtype(prevRefSmooth) != 0 ? 0 : prevRefSmooth
	variable prevRefIsSame = abs(s.refSmooth - prevRefSmooth) < 10^-9
	s.UIoptions = prevRefIsSame ? s.UIoptions | 2^0	: s.UIoptions								// don't calculate reference smoothing again for performance reasons
	s.UIpinToggle = ParamIsDefault(undoPinning) ? 0 : undoPinning								// this will offset the background in a way which keeps the apparent spectrum position constant
	
	variable csrX, csrY
	variable valMin = FindBaseline(pPath, is2D, csrX, csrY)
if (is2D)
	s.baselineXval	= csrX
	s.baselineYval	= csrY
	s.baselineZval	= valMin
	s.yShift		= Modifiers[1]
	s.valShift		= Modifiers[3]
else
	s.baselineXval	= csrX
	s.baselineYval	= valMin
	s.valShift		= Modifiers[1]
endif
	s.xShift		= Modifiers[0]
	s.valScale		= Modifiers[2]

	int error = doWaveArithmetic(s)																// actually do the comparison
	
	Modifiers[0] = s.xShift																		// write back changes to the parameters
	Modifiers[2] = s.valScale
if (is2D)
	Modifiers[1] = s.yShift
	Modifiers[3] = s.valShift
else
	Modifiers[1] = s.valShift
endif

	//+++++++++++++++++++++++++++++++++ prepare auxillary wave +++++++++++++++++++++++++++++++++++
	Wave/SDFR=stg WaveA
	Duplicate/O s.cmpWave, WaveA;	Note/K WaveA, note(s.cmpWave)
	ControlInfo/W=$pPath popAuxMode;	int auxMode = V_Flag==3 ? WhichListItem(S_value, kAuxListKey) : kAuxMode		// control added in version 4.5
if (is2D)
	Switch(auxMode)
		case 1:
			MatrixFilter/N=(limit(round(sqrt(numpnts(s.cmpWave)*0.001)), 3, 333)) gauss WaveA
		break
		case 2:
			Integrate/DIM=0 WaveA /D=WaveA
//			Integrate/DIM=1 WaveA /D=WaveA
		break
		case 3:
			Differentiate/DIM=0 WaveA /D=WaveA
//			Differentiate/DIM=1 WaveA /D=WaveA
		break
	EndSwitch
	AppendAuxImage(pPath, WaveA, auxMode)
		
	if (!prevRefIsSame)
		CreateRefProfiles(pPath)
	endif
	CreateModProfiles(pPath)
else
	Switch(auxMode)
		case 0:
			MatrixOP/O WaveA = NaN
		break
		case 1:
			Smooth limit(round(DimSize(s.cmpWave,0)*0.05), 1, 2^15-1), WaveA
		break
		case 2:
			Integrate WaveA /D=WaveA
		break
		case 3:
			Differentiate WaveA /D=WaveA
		break
	EndSwitch
	ModifyGraph/Z/W=$getGraph(pPath) muloffset(WaveA)={0,0}
endif
	
	ImageSetRange(pPath)
	UpdateGraphLegend(pPath, err=error)
	
	if (debugGetTiming)	
		Print "Result calculation took", StopMSTimer(timerRef)/1000, "microseconds"
	endif
	return 0
end
//+++++++++++++++++++++++++++++++++ which baseline value to use ++++++++++++++++++++++++++++++++++
static function FindBaseline(string pPath, int is2D, variable &csrX, variable &csrY)
	string gName = getGraph(pPath)
	string csrList = "A;B;C;D;E;F;G;H;"															// I and J are reserved for ROI
	
	variable i, minV = inf
	csrX = NaN; csrY = NaN
	for (i = 0; i < ItemsInList(csrList); i++)
		string csr = StringFromList(i,csrList)
		if (!StringMatch(CsrWave($csr, gName),"WaveR"))
			continue
		endif
		variable curr = is2D ? zcsr($csr, gName) : vcsr($csr, gName)
		if (curr < minV)
			if (is2D)
				csrX = xcsr(A, gName); csrY = vcsr(A, gName); minV = zcsr(A, gName)
			else
				csrX = xcsr(A, gName); minV = vcsr(A, gName)
			endif
		endif
	endfor
	
	SVar/SDFR=getWorkFolder(pPath) locOriginal
	return numType(minV) != 0 ? Wavemin($locOriginal) : minV									// use either the wave minimum or cursor location
end

//################################################################################################

static function AppendAuxImage(string pPath, wave aux, int auxMode)
	string gName = getGraph(pPath)
	string wName = NameOfWave(aux)
	if (auxMode > 0)
		if (FindListItem(wName,ImageNameList(gName,";"))<0)
			AppendImage/W=$gName/L=Ldiff aux
			ModifyImage/W=$gName $wName ctab= {*,*,Terrain,0}
		endif
	else
		RemoveImage/W=$gName/Z $wName
	endif
end

//################################################################################################

static function CreateRefProfiles(string pPath)
	DFREF stg = getWorkFolder(pPath)
	Wave/Z/SDFR=stg WaveR
	if (!WaveExists(WaveR))
		return -1
	endif
	if (WaveDims(WaveR)==2)
		doProfiles(getGraph(pPath), stg, "WaveR")
	endif
	return 0
end

static function CreateModProfiles(string pPath)
	DFREF stg = getWorkFolder(pPath)
	Wave/SDFR=stg WaveO
	if (WaveDims(WaveO)==2)
		doProfiles(getGraph(pPath), stg, "WaveO")
		doProfiles(getGraph(pPath), stg, "WaveD")
	endif
	return 0
end

//++++++++++++++++++++++++++++++++++ create profile helper +++++++++++++++++++++++++++++++++++++++
static function doProfiles(string gName, DFREF stg, string wName)
	int doCsr = strlen(CsrInfo(I,gName)) > 0 && strlen(CsrInfo(J,gName)) > 0
	Wave work = stg:$wName
	if (doCsr)
		Duplicate/Free/RMD=[][qcsr(I,gName),qcsr(J,gName)] work, temp; Wave work = temp
	endif
	MatrixOP/O stg:$(wName+"_prH") = sumRows(work)/numCols(work)
	SetScale/P x, DimOffset(work,0), DimDelta(work,0), WaveUnits(work,0), stg:$(wName+"_prH")
	Wave work = stg:$wName
	if (doCsr)
		Duplicate/Free/RMD=[pcsr(I,gName),pcsr(J,gName)][] work, temp; Wave work = temp
	endif
	MatrixOP/O stg:$(wName+"_prV") = sumRows(work^t)/numRows(work)
	SetScale/P x, DimOffset(work,1), DimDelta(work,1), WaveUnits(work,1), stg:$(wName+"_prV")
end

//################################################################################################

function DiffGen_ProcessWindowEvent(STRUCT WMWinHookStruct &s)
	string gName = getGraph(s.winName)															// can be called from AskQuitPanel as well => construct panel path manually
	string pPath = getPanel(gName)
	DFREF stg = getWorkFolder(pPath)
	
	variable axDivider, dragAxis = 0															// for a dynamic divider line
	if (s.EventCode > 2 && s.EventCode < 6 && CmpStr(gName,s.winName) == 0)						// for mouse events
		sscanf StringByKey("axisEnab(x)", AxisInfo(gName,"left"), "="), "{%f,", axDivider
		axDivider -= 0.01																		// center split position
		GetWindow $gName, psizeDC
		variable mouseClose = abs((s.mouseloc.v-v_top)-(1-axDivider)*(v_bottom-v_top)) < 5
		variable rMousePos	= 1-(s.mouseloc.v-v_top)/(v_bottom-v_top)							// relative mouse position on the plot
	endif
	
	int HookTakeover = 0
	StrSwitch (s.eventName)
		case "killVote":
			if (WinType(s.winName) == 1)														// comes from the main graph (there should not be a kill event from a panel anyway)
				HookTakeover = 1
				SVar/SDFR=stg locOriginal
				Wave/SDFR=stg WaveO
				Wave   srcwave = $(locOriginal)
				Wave/Z modwave = $(ParseFilePath(1, locOriginal, ":", 1, 0) + PossiblyQuoteName(NameOfWave(srcwave) + kModEnd))
				
				int changed = 0
				if (WaveExists(modwave))
					changed = !EqualWaves(WaveO, modwave, 5)
				else
					changed = !EqualWaves(WaveO, srcwave, 5)
				endif
				if (changed)
					if (!Wintype("#AskQuitPanel"))												// this will create a little panel in case of changes
						variable pixConv = (CmpStr(IgorInfo(2), "Windows") == 0 && ScreenResolution == 96)? 72/ScreenResolution : 1
						variable xCenter = (s.WinRect.right - s.WinRect.left)/2*72/ScreenResolution, yCenter = (s.WinRect.bottom - s.WinRect.top)/2*72/ScreenResolution
						NewPanel/N=AskQuitPanel/HOST=$s.WinName/W=(xCenter-105*pixConv,yCenter-38*pixConv,xCenter+105*pixConv,yCenter+38*pixConv)
						DrawText/W=$(s.winName+"#AskQuitPanel") 20,28,"Quiting: Save made changes?"
						Button bProgSaveQuit	,win=$(s.winName+"#AskQuitPanel")	,pos={20,33}	,size={50,30}	,title="\f01Save"	,proc=DiffGen_ButtonControl	,help={"Save changes and quit. Hold the 'alt/option' key to save the difference spectrum with the name of the reference wave instead of the editing wave."}
						Button bProgUndoQuit	,win=$(s.winName+"#AskQuitPanel")	,pos={80,33}	,size={50,30}	,title="Discard"	,proc=DiffGen_ButtonControl	,help={"Discard changes and quit."}
						Button bProgCancel		,win=$(s.winName+"#AskQuitPanel")	,pos={140,33}	,size={50,30}	,title="Cancel"		,proc=DiffGen_ButtonControl
						if (CmpStr(IgorInfo(2), "Macintosh") == 0)
							ModifyControlList ControlNameList(s.winName+"#AskQuitPanel",";","*") fsize=10
						endif
						SetActiveSubwindow $s.WinName											// restore focus to main window
					endif
				else
					ExecuteProgramButtons(pPath, "bProgUndoQuit")
				endif
			endif
		break
		case "keyboard":
			string key = ""
			Switch(s.keycode)
				case 11:
					key = "bStretch"
					break
				case 12:
					key = "bShrink"
					break
				case 28:
					key = "bLeft"
					break
				case 29:
					key = "bRight"
					break
				case 30:
					key = "bUp"
					break
				case 31:
					key = "bDown"
					break
				case 42:	// * key
					key = "setMode_3"
					break
				case 43:	// + key
					key = "setMode_2"
					break
				case 45:	// - key
					key = "setMode_1"
					break
				case 47:	// / key
					key = "setMode_4"
					break
				case 105:	// i key
					ControlInfo/W=$pPath InvertMode
					CheckBox InvertMode ,win=$pPath ,value=1-V_Value
					break
				case 112:	// p key
					key = "cPin"
					break	
				case 114:	// r key
					key = "bUndo"
					break
				case 115:
					ControlInfo/W=$pPath ref2D_DoShow
					if (V_flag == 2)
						CheckBox ref2D_DoShow ,win=$pPath ,value=1-V_Value
						ImageSetAlpha(pPath)
					endif
					break
				case 1:		// home key
					key = "smthInc"
					break	
				case 4:		// end key
					key = "smthDec"
					break
				case 0:		// insert key
					key = "bZUp"
					break
				case 127:	// delete key
					key = "bZDown"
					break
			EndSwitch
			
			variable ignore = FindListItem(key,"bLeft;bRight;bUp;bDown;")>-1 && (GetKeyState(0) & 2^1)	// move cursors with alt key
			int hasCsrA = strlen(CsrInfo(A,gName)) && strsearch(CsrInfo(A,gName),"/A=0",0,2)==-1
			int hasCsrB = strlen(CsrInfo(B,gName)) && strsearch(CsrInfo(B,gName),"/A=0",0,2)==-1
			int hasCsrI = strlen(CsrInfo(I,gName)) && strsearch(CsrInfo(I,gName),"/A=0",0,2)==-1
			int hasCsrJ = strlen(CsrInfo(J,gName)) && strsearch(CsrInfo(J,gName),"/A=0",0,2)==-1
			if ((hasCsrA || hasCsrB || hasCsrI || hasCsrJ) && ignore)
				Execute/Q/P "WaveArthimeticTool#CreateRefProfiles(\""+pPath+"\")"				// needs to be done after cursor movement has finished
				Execute/Q/P "WaveArthimeticTool#CreateModProfiles(\""+pPath+"\")"
				Execute/Q/P "WaveArthimeticTool#UpdateGraphLegend(\""+pPath+"\")"
				break
			endif
			
			HookTakeover = 1
			ProcessKeyEvents(pPath, Key)														// do the modification according to the pressed key
		break
		case "mousedown":
			if (mouseClose)																		// initialize axis drag state
				SetWindow $pPath userdata(axdragOn)="1"
				HookTakeover = 1
			endif
		break
		case "mouseup":
				SetWindow $pPath userdata(axdragOn)="0"											// release axis drag
		break
		case "mousemoved":
			//++++++++++++++++++++++++++++++++ axis quick drag ++++++++++++++++++++++++++++++++++
			if (!WinType(pPath))																// catches slow window buildup in Igor 6
				break
			endif
			dragAxis = str2num(GetUserData(pPath, "", "axdragOn"))
			dragAxis = numtype(dragAxis) != 0 ? 0 : dragAxis
			if (dragAxis)
                GetMouse/W=$gName
                if (!(V_flag & 1))																// mouse button released? => maybe we missed a mouse-up event?
                    SetWindow $pPath userdata(axdragOn)="0"
                    dragAxis = 0
                endif
            endif
			DrawAction/W=$gName getgroup=axisSplitLine, delete									// draws and sets divider line for dynamic axis drag
			if (dragAxis)
				axDivider = limit(rMousePos, 0.15, 0.85)										// set new divider
				ModifyGraph/W=$gName axisEnab(left)	 = {axDivider + 0.01, 1}
				ModifyGraph/W=$gName axisEnab(Ldiff) = {0, axDivider - 0.01}
				if (FindListItem("MainColorDisplay", AnnotationList(gName)) > -1)				// scale image color legend
					ColorScale/W=$gName/C/N=MainColorDisplay heightPct=100-(axDivider+0.01)*100
					ColorScale/W=$gName/C/N=DiffColorDisplay heightPct=(axDivider-0.01)*100
				endif
				HookTakeover = 1
			endif
			if (dragAxis || mouseClose)															// draw a line for visualization
				SetDrawEnv/W=$gName linefgc=(36000,15000,59000) ,dash=2 ,linethick= 1.50
				SetDrawEnv/W=$gName gstart, gname=axisSplitLine
				DrawLine/W=$gName -0.1, 1-axDivider, 1.1, 1-axDivider
				SetDrawEnv/W=$gName gstop, gname=axisSplitLine
				s.cursorCode = 6																// change mouse cursor symbol
				s.doSetCursor = 1
			endif
		break
		case "cursormoved":	
			CreateRefProfiles(pPath)
			CreateModProfiles(pPath)
			UpdateGraphLegend(pPath)
		break
	EndSwitch
	return HookTakeover
end

//################################################################################################

static function/S LoadWavesAndSettings(string pPath)											// creates or copies the working waves from the originals
	DFREF stg = getWorkFolder(pPath)
	SVar/SDFR=stg locOriginal
	Wave/SDFR=stg Modifiers																		// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplier, [3] = z offset (2D data)
	string srcname = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	string scrpath = ParseFilePath(1, locOriginal, ":", 1, 0)
	Wave   srcwave = $(locOriginal)
	Wave/Z modwave = $(scrpath + PossiblyQuoteName(srcname + kModEnd))
	
	if (DimSize(Modifiers,0) < 4)
		Redimension/N=(4) Modifiers
	endif
	
	string recreate = "", outnotes = "", value = ""
	if (WaveExists(modwave))																	// duplicate pre-existing mod wave into temp folder
		Duplicate/O modwave stg:WaveO
		recreate = note(modwave)																// extract modify parameters from wave note
	else
		Duplicate/O srcwave stg:WaveO
		Modifiers = 0; Modifiers[2] = 1															// reset modify parameters
	endif
	
	Duplicate/O stg:WaveO stg:WaveD, stg:WaveA
	Wave/Z/SDFR=stg WaveO, WaveR, WaveD, WaveA
	WaveD = NaN; WaveA = NaN																	// set to NaNs at first to render the wave invisible
	if (!WaveExists(WaveR))
		Duplicate WaveO stg:WaveR;	Wave/SDFR=stg WaveR;	WaveR = NaN							// create a dummy reference wave
	endif
	
	int i, is2D = WaveDims(WaveO)==2
	for (i = 0; i < ItemsInList(kOutEnd); i++)
		Wave/Z outwave= $(scrpath + PossiblyQuoteName(srcname + StringFromList(i,kOutEnd)))
		if (WaveExists(outwave))
			outnotes = note(outwave)
			// *********************** Compatibility code: load old diff waves *******************
			variable oldDiff = StringMatch(outnotes,"*Difference created from*") + StringMatch(outnotes,"*difference spectrum to*")
			// ***********************************************************************************
			if (!WaveExists(modwave) || oldDiff)												// if no modification wave exits try to load settings from outnote
				recreate = outnotes
			elseif (StringMatch(outnotes,"*"+kRefNameKey+"*"))									// the diff wave was saved with the reference name => switch edit and ref wave
				recreate += "/r"+kRefNameKey
			endif
			break
		endif
	endfor
	
	// ++++++++++++++++++++++ load saved control settings from wave note +++++++++++++++++++++++++
	variable srcsmth = 0, refsmth = 0, from = 0, xVal, yVal
	if (strlen(recreate))
		from = strsearch(recreate,"Created from:",0,2);
		from = from == -1 ? 0 : from
		recreate = recreate[from,inf]
		
		Modifiers[2] = str2num(StringByKey(kScaleKey, recreate, "=", "\r"))
		value = StringByKey(kShiftKey, recreate, "=", "\r")
		if (strsearch(value,"[",0) > -1)
			sscanf value, "[%f,%f]", xVal, yVal
			Modifiers[0] = xVal
			Modifiers[1] = yVal
			Modifiers[3] = str2num(StringByKey(kOffsetKey,recreate, "=", "\r"))
		else
			Modifiers[0] = str2num(StringByKey(kShiftKey, recreate, "=", "\r"))
			Modifiers[1] = str2num(StringByKey(kOffsetKey,recreate, "=", "\r"))
		endif
		
		srcsmth = str2num(StringByKey(kModSmthKey,	recreate, "=", "\r"))
		refsmth	= str2num(StringByKey(kRefSmthKey,	recreate, "=", "\r"))
		srcsmth = numtype(srcsmth) == 0 ? srcsmth : 0
		refsmth = numtype(refsmth) == 0 ? refsmth : 0
		// ************* Compatibility code: extract mod settings from wave directly *************
			if (numtype(sum(Modifiers)) != 0 && WaveExists(modwave))
				Modifiers[0] = ChangePrecision(DimOffset(modwave,0) - DimOffset(srcwave,0),5)
				Modifiers[1] = ChangePrecision(Wavemin(modwave) - Wavemin(srcwave),5)
				Modifiers[2] = round(( (Wavemax(modwave) - Wavemin(modwave)) / (Wavemax(srcwave) - Wavemin(srcwave)))*100000 ) / 100000
			endif
			if (numtype(sum(Modifiers)) != 0)
				Modifiers = 0; Modifiers[2] = 1													// no useful modifiers found
			endif
		// ***************************************************************************************
	endif
	
	string xUnit = WaveUnits(srcwave,0); xUnit[0] = SelectString(strlen(xUnit),""," ")
	string yUnit = WaveUnits(srcwave,1); yUnit[0] = SelectString(strlen(yUnit),""," ")
	variable initX = abs(ChangePrecision(DimDelta(srcwave,0),2))
	variable initY = abs(ChangePrecision(DimDelta(srcwave,1),2))
	variable initZ = abs(ChangePrecision(Wavemax(srcwave)*0.001,1))								// initial offset value (0.1% of max Y value)

if(is2D)
	Setvariable Offset		,win=$pPath ,value=_NUM:initY	,limits={initY/50,inf,initY/4}	,format="%g" + yUnit
	Setvariable zOffset		,win=$pPath ,value=_NUM:initZ	,limits={initZ/50,inf,initZ/4}	,format="%g"
else
	Setvariable Offset		,win=$pPath ,value=_NUM:initZ	,limits={initZ/50,inf,initZ/4}	,format="%g"
endif
	Setvariable Shift		,win=$pPath ,value=_NUM:initX	,limits={initX/50,inf,initX/4}	,format="%g" + xUnit
	Setvariable WaveConv	,win=$pPath ,value=_NUM:srcsmth	,limits={0,inf,initX/4}			,format="%g" + xUnit
	Setvariable RefWaveConv	,win=$pPath ,value=_NUM:refsmth	,limits={0,inf,initX/4}			,format="%g" + xUnit
	
	ControlInfo/W=$pPath AbsXshift
	if (abs(V_flag) == 5)																		// for newer versions => 4.4
			Setvariable AbsXshift	,win=$pPath	,limits={-inf,inf,initX/4}	,format="%g" + xUnit
			Setvariable AbsScale	,win=$pPath	,limits={-inf,inf,0.1}
		if (is2D)
			Setvariable AbsYshift	,win=$pPath	,limits={-inf,inf,initY/4}	,format="%g" + yUnit
			Setvariable AbsZshift	,win=$pPath	,limits={-inf,inf,initZ/4}
		else
			Setvariable AbsYshift	,win=$pPath	,limits={-inf,inf,initZ/4}
		endif
	endif
	CheckBox InvertMode	 	,win=$pPath ,value= StringMatch(recreate,"*"+kModeInvKey+"*")
	CheckBox PinBase	 	,win=$pPath ,value=!StringMatch(recreate,"*"+kPinBaseKey+"*")
	
	ControlInfo/W=$pPath ArithMode;		int isModeSelect = (V_flag == 3)						// obsolete since version 5.0
	int setMode = isModeSelect ? V_value-1 : 0													// keep last mode; difference mode (0) is default
	for (i = 0; i < ItemsInList(kSetModeKey); i++)
		setMode = StringMatch(recreate,"*"+StringFromList(i,kSetModeKey)+"*") ? i : setMode
	endfor
	if (isModeSelect)
		PopUpMenu ArithMode	 ,win=$pPath ,mode=(setMode+1)
	endif
	
	string formula = StringByKey(kFormulaKey, recreate, "=", "\r")
	if (!strlen(formula))																		// compatibility with data from before version 5.0
		formula = StringFromList(setMode,kModePresets)
	endif
	ControlInfo/W=$pPath FormulaStr																// new in version 5.0: directly fill in formula
	if (abs(V_flag) == 5)
		Setvariable FormulaStr ,win=$pPath ,value=_STR:formula
	endif
	return recreate																				// copy over to ReloadRefWaveSettings()
end

//################################################################################################

static function ReloadRefWaveSettings(string pPath, string recreate)
	CheckBox AltSaveOption ,win=$pPath ,value=StringMatch(recreate,"*"+kRefNameKey+"*")			// the diff wave was saved with the reference name => switch edit and ref wave
	
	SVar/SDFR=getWorkFolder(pPath) locOriginal
	string oldRef = StringByKey("Reference Wave", recreate,"=","\r")
	string oldSrc = StringByKey("Original Wave", recreate,"=","\r")
	int switched = StringMatch(oldRef, locOriginal)
	
	string curWave = SelectString(switched, oldRef, oldSrc)										// if switched load the original wave into ref. and switch them over later
	string refPath = ParseFilePath(1, curWave, ":", 1, 0)
	string refName = ReplaceString("'",ParseFilePath(0, curWave, ":", 1, 0),"")
	string srcName = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	
	// ****************** Compatibility code: extract wave info from old files *******************
	if (StrLen(refPath) == 0 || !DataFolderExists(refPath))
		if (StringMatch(recreate,"*Difference created from*"))
			refPath = StringByKey("Folder", recreate,":","\r")
			refName = StringByKey("Wave", recreate,":","\r")
			if (StringMatch(srcName,refName))													// don't to load the same wave as both reference and original
				refName = ""
			endif
		elseif(StringMatch(recreate,"*difference spectrum to*"))
			refPath = GetDataFolder(1)
			refName = recreate[StrLen("difference spectrum to "),inf]
			refName = ReplaceString("'",refName, "")
		endif
	endif
	if (StrLen(refPath) == 0 || !DataFolderExists(refPath))										// still not found
		refPath	= GetDataFolder(1)
	endif
	// *******************************************************************************************
	
	DFREF saveDF = GetDataFolderDFR()
	SetDataFolder refPath
	if(Exists(refName) == 1)
		PopUpMenu RefSelect ,win=$pPath ,popmatch=refName										// select the reference wave
		doWaveSelection(pPath, "RefSelect", refName)											// load the old reference wave
		if (switched)
			SwitchSourceAndRefWaves(pPath)
		endif
		
		variable refsmth = str2num(StringByKey(kRefSmthKey, recreate, "=", "\r"))
		refsmth = numtype(refsmth) == 0 ? refsmth : 0
		Setvariable RefWaveConv	,win=$pPath ,value=_NUM:refsmth									// set ref wave convolution control
		
		variable xPos, yPos																		// set cursors to their original position
		string value = StringByKey(kXCsrSetKey, recreate, "=", "\r")
		if (strsearch(value,"[",0) > -1)
			sscanf value, "[%f,%f]", xPos, yPos
			if (numtype(xPos*yPos) == 0)
				Cursor/I/W=$getGraph(pPath) A WaveR xPos, yPos
			endif
		else
			xPos = str2num(value)
			if (numtype(xPos) == 0)
				Cursor/W=$getGraph(pPath) A WaveR xPos
			endif
		endif
	endif
	SetDataFolder saveDF
	return 0
end

//################################################################################################

static function UpdateGraphLegend(string pPath [, int err])
	DFREF stg = getWorkFolder(pPath)
	Wave/SDFR=stg WaveO, WaveR, WaveD, WaveA
	string gName = getGraph(pPath)
	
	UpdateModLegend(stg, gName)
	
	int doCsr = 0, is2D = WaveDims(WaveO)==2
	string text = ""
	WaveStats/Q WaveR
	if (V_numNans < numpnts(WaveR))
		if (is2D)
			int P1, P2, Q1, Q2
			if (strlen(CsrInfo(I,gName)) > 0 && strlen(CsrInfo(J,gName)) > 0)
				sprintf text, "\f02Range:\tX=(%.*g,%.*g), Y=(%.*g,%.*g)\f00\r" 3,xcsr(I,gName),3,xcsr(J,gName),3,vcsr(I,gName),3,vcsr(J,gName)
			endif
			text += "\t\f01min\t\tmax\r"
			[P1, P2, Q1, Q2] = getCursorPoints(WaveO, gName);	WaveStats/Q/RMD=[P1,P2][Q1,Q2] WaveO
			sprintf text,   "%s\K(65535,0,0)\f01Mod.:\f00\K(0,0,0)\t%.*g,\t%.*g" text ,3,V_min,3,V_max
			[P1, P2, Q1, Q2] = getCursorPoints(WaveR, gName);	WaveStats/Q/RMD=[P1,P2][Q1,Q2] WaveR
			sprintf text, "%s\r\K(0,0,65535)\f01Ref.:\f00\K(0,0,0)\t%.*g,\t%.*g" text ,3,V_min,3,V_max
			[P1, P2, Q1, Q2] = getCursorPoints(WaveD, gName);	WaveStats/Q/RMD=[P1,P2][Q1,Q2] WaveD
			sprintf text,					  "%s\r\f01Result\f00:\t%.*g,\t%.*g" text ,3,V_min,3,V_max
		else
			variable X1 = -inf,	X2 = inf
			if (strlen(CsrInfo(A, gName)) > 0 && strlen(CsrInfo(B, gName)) > 0)
				X1 = xcsr(A, gName)
				X2 = xcsr(B, gName)
				doCsr = 1
			endif
			if (strlen(CsrInfo(I, gName)) > 0 && strlen(CsrInfo(J, gName)) > 0)
				X1 = xcsr(I, gName)
				X2 = xcsr(J, gName)
				doCsr = 1
			endif
			if (doCsr)
				sprintf text, "Area between %.1f %s and %.1f %s\r", X1, WaveUnits(WaveO,0), X2, WaveUnits(WaveO,0)
			endif
			MatrixOP/Free tempO = replaceNaNs(WaveO,0); CopyScales WaveO, tempO
			MatrixOP/Free tempR = replaceNaNs(WaveR,0); CopyScales WaveR, tempR
			MatrixOP/Free tempD = replaceNaNs(WaveD,0); CopyScales WaveD, tempD
			variable AreaO = Area(tempO,X1,X2), AreaR = Area(tempR,X1,X2), AreaD = Area(tempD,X1,X2)
			text += "\\s(WaveO) modified\t(area: "		+ num2str(AreaO)
			text += ")\r\\s(WaveR) reference\t(area: "	+ num2str(AreaR) + "; ratio: " + num2str(abs(AreaR / AreaO))
			text += ")\r\\s(WaveD) result\t(area: "		+ num2str(AreaD) + "; ratio: " + num2str(abs(AreaD / AreaO))
			
			ControlInfo/W=$pPath popAuxMode; int auxMode = V_Flag==3 ? WhichListItem(S_value, kAuxListKey) : kAuxMode		// control added in version 4.5
			Switch(auxMode)
				case 1:
					text += ")\r\\s(WaveA) smoothed"
				break
				case 2:
					text += ")\r\\s(WaveA) integrated (area: "	+ num2str(Area(WaveA,X1,X2)) + ")"
				break
				case 3:
					text += ")\r\\s(WaveA) derivative (area: "	+ num2str(Area(WaveA,X1,X2)) + ")"
				break
			EndSwitch
		endif
	endif
	
	if (!ParamIsDefault(err) && err != err_NoError)
		text += "\r"+WaveArithmeticErrorMsg(err)
	endif
	
	Legend/C/N=GraphInfoBox/W=$gName/J/F=0/B=1 text
	if (strlen(text) && is2D)
		Legend/C/N=GraphInfoBox/W=$gName/J/F=2/B=2/T={52,114}									// 2D data => add box
	endif
	
	return 0
end
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
static function [int P1, int P2, int Q1, int Q2] getCursorPoints(wave in, string gName)
	int doCsr = strlen(CsrInfo(I,gName)) > 0 && strlen(CsrInfo(J,gName)) > 0
	P1=0; P2=DimSize(in,0)-1; Q1=0; Q2=DimSize(in,1)-1
	if (doCsr)
		P1=limit(ScaleToIndex(in,xcsr(I,gName),0),0,P2)
		P2=limit(ScaleToIndex(in,xcsr(J,gName),0),0,P2)
		Q1=limit(ScaleToIndex(in,vcsr(I,gName),0),0,Q2)
		Q2=limit(ScaleToIndex(in,vcsr(J,gName),0),0,Q2)
	endif
end

//################################################################################################

static function UpdateModLegend(DFREF stg, string gName)
	Wave/SDFR=stg WaveO, Modifiers																// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplier, [3] = z offset (2D data)
	SVar/SDFR=stg locOriginal
	// ++++++++++++++++++++++++ right legend (modification details) ++++++++++++++++++++++++++++++
	int is2D = WaveDims(WaveO)==2
	string text = "", xUnit = WaveUnits(WaveO,0), , yUnit = WaveUnits(WaveO,1)
	if (abs(Modifiers[0]) > 10^-16)
		text += num2str(Modifiers[0]) + " " + SelectString(strlen(xUnit),"units",xUnit) + SelectString(is2D," shift\r"," x-shift\r")
	endif
	if (abs(Modifiers[1]) > 10^-16 && is2D)
		text += num2str(Modifiers[1]) + " " + SelectString(strlen(yUnit),"units",yUnit) + " y-shift\r"
	endif
	variable offset = WaveMin(WaveO)-WaveMin($locOriginal)
	if (abs(offset) > 10^-10)
		text += num2str(offset) + " offset\r"
	endif
	if (Modifiers[2] != 1)
		text += num2str(abs(Modifiers[2])) + " x scaled\r"
	endif
	variable smth = str2num(StringByKey(kModSmthKey,note(WaveO), "=", "\r"))
	if (numtype(smth) == 0 && smth > DimDelta(WaveO,0)*10^-9)									// avoid rounding problems
		text += num2str(smth) + SelectString(strlen(xUnit),""," ") + xUnit + " smoothed\r"
	endif
	smth = str2num(StringByKey(kRefSmthKey,note(WaveO), "=", "\r"))
	if (numtype(smth) == 0 && smth > DimDelta(WaveO,0)*10^-9)
		text += num2str(smth) + SelectString(strlen(xUnit),""," ") + xUnit + " ref. smooth\r"
	endif
	Legend/C/N=ModificationInfoBox/W=$gName/J/F=0/B=1 "\K(65280,0,0)" + RemoveEnding(text,"\r")
	if (strlen(text) && WaveDims(WaveO)==2)
		Legend/C/N=ModificationInfoBox/W=$gName/J/F=2/B=2										// 2D data => add box
	endif
	return 0
end

//############################### support functions for control elements #########################

function DiffGen_ModeUpdater(STRUCT WMCheckboxAction& s) : CheckBoxControl						// only for backwards compatibility
	if(s.eventCode == 2)
		CalculateResult(s.win, undoPinning=(1-2*s.checked)*StringMatch(s.ctrlName,"PinBase"))
	endif
	return 0
end

//################################################################################################

function DiffGen_PinBaseMode(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if(s.eventCode == 2)
		CalculateResult(s.win, undoPinning=(1-2*s.checked))										// undo the pinning state before doing the calculation
	endif
	return 0
end

//################################################################################################

function DiffGen_InvertMode(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if(s.eventCode == 2)
		CalculateResult(s.win)
	endif
	return 0
end

//################################################################################################

function DiffGen_WaveConvolve(STRUCT WMSetVariableAction &s) : SetVariableControl
	if (s.eventCode == 1 || s.eventCode == 2)
		CalculateResult(s.win)
	endif
	return 0
end

//################################################################################################

function DiffGen_ModeSelect(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		if (!CmpStr(s.ctrlName,"popFormula"))
			Setvariable FormulaStr ,win=$s.win ,value=_STR:s.popStr								// new in version 5.0: fill in expression
		endif
		CalculateResult(s.win)
		ModifyControl bUndo ,win=$s.win ,activate												// move focus away from the popup (prevents accidental selection of a different wave with the cursor keys)
	endif
	return 0
end

//################################################################################################

function ImageColorProc(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		if (!CmpStr(s.ctrlName,"ref2D_Color"))
			DFREF stg = getWorkFolder(s.win)
			ColorTab2Wave $s.popStr; Wave M_colors;	Redimension/N=(-1,4) M_colors
			Duplicate/O M_colors, stg:refColors; 	KillWaves/Z M_colors
			ImageSetAlpha(s.win)
		endif
		ImageSetRange(s.win)
	endif
	return 0
end

//################################################################################################

Function AlphaSliderProc(STRUCT WMSliderAction &s) : SliderControl
	if (s.eventCode & 1)
		ImageSetAlpha(s.win)
	endif
	return 0
End

//################################################################################################

function ToggleRefVisibility(STRUCT WMCheckboxAction& s) : CheckBoxControl
	if(s.eventCode == 2)
		ImageSetAlpha(s.win)
	endif
	return 0
end

//################################################################################################

function RangeSliderProc(STRUCT WMSliderAction &s) : SliderControl
	if (s.eventCode & 1)
		ImageSetRange(s.win)
	endif
	return 0
End

//################################################################################################

static function ImageSetAlpha(string pPath)
	wave/Z/SDFR=getWorkFolder(pPath) refColors
	if (WaveExists(refColors))
		ControlInfo/W=$pPath ref2D_Alpha;	variable alpha = V_Value
		ControlInfo/W=$pPath ref2D_DoShow;	variable show = V_Value
		refColors[][3] = 65535*alpha*show
	endif
end

//################################################################################################

static function ImageSetRange(string pPath)
	DFREF stg = getWorkFolder(pPath)
	wave/Z/SDFR=stg WaveO, WaveR, refColors
	if (!WaveExists(refColors))
		return -1
	endif
	WaveStats/Q WaveO;	variable minO = V_min, maxO = V_max
	WaveStats/Q WaveR;	variable minR = V_min, maxR = V_max	
	variable minVal = numtype(minR) == 0 ? min(minR,minO) : minO
	variable maxVal = numtype(maxR) == 0 ? max(maxR,maxO) : maxO
	ControlInfo/W=$pPath ValRangeLow;	variable minRange = V_Value
	ControlInfo/W=$pPath ValRangeHigh;	variable maxRange = V_Value
	ControlInfo/W=$pPath src2D_Color;	string color = S_Value
		ModifyImage/W=$getGraph(pPath) WaveO ctab={minVal+(maxVal-minVal)*minRange,maxVal*maxRange,$color,0}
	if(FindListItem("WaveR",ImageNameList(getGraph(pPath),";")) > -1)
		ModifyImage/W=$getGraph(pPath) WaveR ctab={minVal+(maxVal-minVal)*minRange,maxVal*maxRange,stg:refColors,0}
	endif
end

//################################################################################################

function DiffGen_PopupWaveSelect(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)
		doWaveSelection(s.win, s.ctrlName, s.popStr)
		ModifyControl bUndo ,win=$s.win ,activate												// move focus away from the popup (prevents accidental selection of a different wave with the cursor keys)
	endif
	return 0
end
//+++++++++++++++++++++++++++++++++++++ popup help function +++++++++++++++++++++++++++++++++++++
function/S DiffGen_CreateWaveList(string type)													// backwards compatible => changed in version 4.6
	return getPopWaveList(StringMatch(type,"Ref"), 1)
end

function/S getPopWaveList(int which, int dim)
	string pPath = getPanel(StringFromList(0,WinList("DiffGen*", ";","WIN:1")))					// have to assume the popup is invoked while the attached graph is on top
	if (strlen(pPath))
		SVar/SDFR=getWorkFolder(pPath) locOriginal
		string name	 = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
		string waves = SortList(WaveList("*",";","DIMS:"+num2str(dim)+",TEXT:0"), ";", 16)
		string presets = ReplaceString(" ",kModePresets,"")
		ControlInfo/W=$pPath ArithMode;  int setMode = V_flag==3 ? V_value : 0
		ControlInfo/W=$pPath FormulaStr; 	 setMode = V_flag==5 ? WhichListItem(ReplaceString(" ",S_value,""), presets)+1 : setMode
		ControlInfo/W=$pPath popAuxMode; int auxMode = V_Flag==3 ? WhichListItem(S_value, kAuxListKey) : kAuxMode		// control added in version 4.5	
		waves = RemoveFromList(name+StringFromList(setMode,kOutEnd)+";"+name+kModEnd+";",waves)	// remove the associated result waves (cannot remove reference name here or switching will not work)
		if (auxMode > 0)
			waves = RemoveFromList(name+StringFromList(auxMode-1,kAuxEnd),waves)
		endif
		return SelectString(which,waves,"none;" + RemoveFromList(name,waves))					// in case of reference wave popup - create wave list from current folder without original wave
	else
		return ""
	endif
end

//################################################################################################

static function doWaveSelection(string pPath, string type, string name)
	DFREF stg = getWorkFolder(pPath)
	SVar/SDFR=stg locOriginal, locReference
	
	string gName = getGraph(pPath)
	StrSwitch (type)
		case "OrigSelect":
			string oldName = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
			locOriginal = GetWavesDataFolder($name, 2)
			LoadWavesAndSettings(pPath)
			GetWindow  $gName, wtitle
			DoWindow/T $gName, ReplaceString(oldName,S_value,"") + name
		break
		case "RefSelect":
			Wave/SDFR=stg WaveR
			locReference = ""; WaveR = NaN
			int is2D = WaveDims(WaveR)==2
			if (is2D)
				RemoveImage/W=$gName/Z WaveR
				Wave/Z/SDFR=stg WaveR_prV, WaveR_prH
				if (WaveExists(WaveR_prV) && WaveExists(WaveR_prH))
					WaveR_prH = NaN; CopyScales stg:WaveO_prH, WaveR_prH
					WaveR_prV = NaN; CopyScales stg:WaveO_prV, WaveR_prV
				endif
			else
				RemoveFromGraph/W=$gName/Z WaveR
			endif
			if (!StringMatch(name,"none"))
				locReference = GetWavesDataFolder($name, 2)
				Duplicate/O $name stg:WaveR
				if (is2D)
					CreateRefProfiles(pPath)
					AppendImage/W=$gName WaveR
					CheckBox ref2D_DoShow ,win=$pPath ,value=1
					ImageSetRange(pPath)
					ImageSetAlpha(pPath)
				else
					AppendToGraph/W=$gName/C=(0,0,65280) WaveR
				endif
			endif
		break
	endswitch
	CalculateResult(pPath)
	return 0
end

//################################################################################################

static function SwitchSourceAndRefWaves(string pPath)
	SVar/SDFR=getWorkFolder(pPath) locOriginal, locReference
	Wave   scrWave = $locOriginal
	Wave/Z refWave = $locReference
	
	DFREF sav = GetDataFolderDFR()
	if (WaveExists(RefWave))
		string scrName = NameOfWave(scrWave)
		string refName = NameOfWave(refWave)
		
		SetDataFolder GetWavesDataFolderDFR(refWave)
		PopUpMenu OrigSelect ,win=$pPath ,popMatch=refName
		doWaveSelection(pPath, "OrigSelect", refName)
		
		SetDataFolder GetWavesDataFolderDFR(scrWave)
		PopUpMenu RefSelect	,win=$pPath ,popMatch=scrName
		doWaveSelection(pPath, "RefSelect", scrName)
	endif
	SetDataFolder sav
	return 0
end

//################################################################################################

function DiffGen_ButtonControl(STRUCT WMButtonAction &s) : ButtonControl
	if (s.eventCode == 2)
		if (StringMatch(s.ctrlName,"bProg*"))													// these buttons have a different execute function (no difference calculation)
			ExecuteProgramButtons(s.win, s.ctrlName)
		else
			ProcessKeyEvents(s.win, s.ctrlName)
		endif
	endif
	return 0
end
//+++++++++++++++++++++++++++++++++++++ button help function +++++++++++++++++++++++++++++++++++++
static function ExecuteProgramButtons(string inPath, string type)
	string gName = getGraph(inPath)																// can be called from AskQuitPanel as well => construct panel path manually
	string pPath = getPanel(gName)
	DFREF stg = getWorkFolder(pPath)
	StrSwitch(type)
		case "bProgSaveQuit":
			if (WinType(gName+"#AskQuitPanel") == 7)
				KillWindow $(gName+"#AskQuitPanel")
			endif
			SaveAndGraphResult(pPath)															// no break here => also kill window										
			if ( (GetKeyState(0) & 2^0) != 0 )													// do not quit when ctrl is pressed
				break
			endif
		case "bProgUndoQuit":
			SetWindow $pPath hook(DiffWindow)=$""												// release hook to not trigger any events
			SetWindow $gName hook(DiffWindow)=$""
			KillWindow $gName																	// really kill the window, before deleting everything
			if (CountObjectsDFR(stg, 4) == 0)													// makes sure to not kill the root folder with other open instances
				KillDataFolder/Z stg
			endif
			DFREF pkroot = getWorkFolder("")
			variable Objects = CountObjectsDFR(pkroot, 1)+CountObjectsDFR(pkroot, 2)+CountObjectsDFR(pkroot, 3)+CountObjectsDFR(pkroot, 4)
			if (Objects == 0 && DataFolderRefStatus(pkroot) != 0)
				KillDataFolder pkroot
			endif
			DFREF pkg = root:Packages:
			Objects = CountObjectsDFR(pkg, 1)+CountObjectsDFR(pkg, 2)+CountObjectsDFR(pkg, 3)+CountObjectsDFR(pkg, 4)
			if (Objects == 0)																	// clean up Packages folder if empty
				KillDataFolder pkg
			endif
		break
		case "bProgCancel":																		// just remove the sub-panel
			KillWindow #AskQuitPanel
		break
		case "bProgSwitch":
			SwitchSourceAndRefWaves(pPath)
		break
		case "bProgActivateROI":
			int doCsr = strlen(CsrInfo(I,gName)) > 0 && strlen(CsrInfo(J,gName)) > 0
			GetMarquee/W=$gName bottom, left
			string option = SelectString(doCsr && !V_Flag, "Set", "Clear")+" Region of Interest"
			MarqueeHandler(gName, option)
		break
		case "bProgShowProfile":
			ShowProfilePanel(stg, gName)
		break
	EndSwitch
	return 0
end
//+++++++++++++++++++++++++++++++++++++ button help function +++++++++++++++++++++++++++++++++++++
static function ProcessKeyEvents(string pPath, string ctrlName)									// update modification parameters and recalculate difference
	DFREF stg = getWorkFolder(pPath)
	SVar/SDFR=stg locOriginal
	Wave/SDFR=stg Modifiers																		// Modifiers: [0] = horizontal shift, [1] = vertical shift, [2] = multiplier, [3] = z offset (2D data)
	Wave srcwave = $(locOriginal)
	int is2D = WaveDims(srcWave)==2
	
	variable ctrlKey = (GetKeyState(0) & 2^0) != 0
	variable altKey  = (GetKeyState(0) & 2^1) != 0
	variable shftKey = (GetKeyState(0) & 2^2) != 0

	variable MinScaleLimit = 0.0001																// lower limit for scaling (multiplication)
	variable Accelerate = 1, Rev = 1, newVal = 0												// acceleration modifier & reverse action modifier (to trigger the opposite effect)
	Accelerate = shftKey ? 10 : Accelerate														// when additional keys are pressed, modify the experience
	Accelerate = ctrlKey ? 0.5 : Accelerate														// gets faster (shift = 10x) or slower (ctrl = 1/2)
	Accelerate = ctrlKey && shftKey ? 0.1 : Accelerate											// super slow (1/10) with both shift and ctrl
	
	if (DimSize(Modifiers,0) < 4)
		Redimension/N=(4) Modifiers
	endif
	
	Strswitch (ctrlName)
		case "bZDown":
			Rev = -1
		case "bZUp":
			ControlInfo/W=$pPath zOffset
			Modifiers[3] += V_Value * Accelerate * Rev
			Modifiers[3] = abs(Modifiers[3]) < 10^-16 ? 0 : Modifiers[3]						// get rid of floating point errors
		break
		case "bDown":
			Rev = -1
		case "bUp":
			ControlInfo/W=$pPath Offset
			Modifiers[1] += V_Value * Accelerate * Rev
			Modifiers[1] = abs(Modifiers[1]) < 10^-16 ? 0 : Modifiers[1]
		break
		case "bLeft":
			Rev = -1
		case "bRight":
			ControlInfo/W=$pPath Shift
			Modifiers[0] += V_Value * Accelerate * Rev
			Modifiers[0] = abs(Modifiers[0]) < 10^-16 ? 0 : Modifiers[0]
		break
		case "bStretch":
			Rev = -1
		case "bShrink":
			ControlInfo/W=$pPath Multiplier
			Modifiers[2]  = (Modifiers[2] == MinScaleLimit && Rev == -1) ? 0 : Modifiers[2]		// scroll away smoothly from minimum limit
			Modifiers[2] -= (V_Value * Accelerate * Rev)/100									// do the modification
			Modifiers[2]  = Modifiers[2] < MinScaleLimit ? MinScaleLimit : Modifiers[2]			// make sure not to shrink the data to zero or invert
		break
		case "bUndo":																			// resets the modifications by overwriting with original data
			Modifiers = 0; Modifiers[2] = 1
			Setvariable WaveConv	,win=$pPath ,value=_NUM:0
			Setvariable RefWaveConv	,win=$pPath ,value=_NUM:0
			if((GetKeyState(0) & 2^0) != 0)														// also reset reference wave if ctrl key is pressed
				PopUpMenu RefSelect ,win=$pPath ,mode=1
				doWaveSelection(pPath, "RefSelect","none")
			endif
		break
		case "bUndoShift":
			Modifiers[0] = 0
		break
		case "bUndoOffset":
			Modifiers[1] = 0
		break
		case "bUndoScale":
			Modifiers[2] = 1
		break
		case "bUndoZOffset":
			Modifiers[3] = 0
		break
		case "cPin":
			ControlInfo/W=$pPath PinBase
			CheckBox PinBase ,win=$pPath ,value=1-V_Value
			Rev = V_Value ? 1 : -1
			variable csrX, csrY
			variable valMin = FindBaseline(pPath, is2D, csrX, csrY)
			Modifiers[1+2*is2D] += Rev*valMin*(1 - Modifiers[2])								// add or subtract the scaled amount to the offset (prevents 'snapping')
		break
		case "smthDec":
			Rev = -1
		case "smthInc":
			string which = SelectString(altKey,"WaveConv","RefWaveConv")						// alt key modifies reference
			ControlInfo/W=$pPath $which
			newVal = V_Value + Rev * abs(ChangePrecision(DimDelta(srcwave,0),2))/4 * Accelerate
			newVal = newVal < 0 ? 0 : newVal
			Setvariable $which ,win=$pPath ,value=_NUM:newVal
		break
		default:
			if (StringMatch(ctrlName,"setMode_*"))
				newVal = str2num(ReplaceString("setMode_",ctrlName,""))
				ControlInfo/W=$pPath ArithMode													// obsolete in version 5.0
				if (abs(V_flag) == 3)
					PopUpMenu ArithMode ,win=$pPath ,mode=newVal
				endif
				ControlInfo/W=$pPath FormulaStr													// new in version 5.0: directly fill in formula
				if (abs(V_flag) == 5)
					Setvariable FormulaStr ,win=$pPath ,value=_STR:StringFromList(newVal-1,kModePresets)
				endif
			endif
		break
	Endswitch
	
	CalculateResult(pPath)
	return 0
end

//################################################################################################

static function SaveAndGraphResult(string pPath)
	DFREF stg = getWorkFolder(pPath)
	SVar/SDFR=stg locOriginal, locReference
	Wave/SDFR=stg WaveO, WaveD, WaveA
	
	string presets = ReplaceString(" ",kModePresets,"")
	ControlInfo/W=$pPath ArithMode;  	int setMode = V_flag==3 ? V_value : 0
	ControlInfo/W=$pPath FormulaStr;		setMode = V_flag==5 ? WhichListItem(ReplaceString(" ",S_value,""), presets)+1 : setMode
	ControlInfo/W=$pPath popAuxMode;	int auxMode = V_Flag==3 ? WhichListItem(S_value, kAuxListKey) : kAuxMode	// control added in version 4.5
	ControlInfo/W=$pPath cAuxSave;		int saveAux = V_Flag==2 ? V_Value : kSaveAux								// control added in version 4.5
	ControlInfo/W=$pPath AltSaveOption;	int refSave = V_flag==2 ? V_value : 0
	variable altKey = (GetKeyState(0) & 2^1) != 0												// keyboard shortcut: Press alt to write the difference wave with the reference wave as name
	int doGraph = 0, i, is2D = WaveDims(WaveO)==2
	
	string savedir = ParseFilePath(1, locOriginal, ":", 1, 0)
	string srcname = ReplaceString("'",ParseFilePath(0, locOriginal, ":", 1, 0),"")
	string modname = PossiblyQuoteName(srcname + kModEnd)
	string difname = PossiblyQuoteName(srcname + StringFromList(setMode,kOutEnd))
	string auxname = PossiblyQuoteName(srcname + StringFromList(auxMode-1,kAuxEnd))
	string srcname_copy = srcname
	Wave   srcwave = $(locOriginal)
	Wave/Z refwave = $(locReference)
	if ((altKey || refSave)	&& WaveExists(refwave))
		savedir	= ParseFilePath(1, locReference, ":", 1, 0)										// change to ref folder
		difname = PossiblyQuoteName(NameOfWave(refwave) + StringFromList(setMode,kOutEnd))
		auxname	= PossiblyQuoteName(NameOfWave(refwave) + StringFromList(auxMode-1,kAuxEnd))
		Note WaveO, "\r"+kRefNameKey
		Note WaveD, "\r"+kRefNameKey
	endif
	
	modname = SelectString(CmpStr(modname,srcname),"noname_err",modname)						// make sure these never overlap
	difname = SelectString(CmpStr(difname,srcname),"noname_err",difname)
	auxname = SelectString(CmpStr(auxname,srcname),"noname_err",auxname)
	
	KillWaves/Z $(savedir + modname)															// cleanup of old waves
	KillWaves/Z $(savedir + difname)
if (auxMode)
	KillWaves/Z $(savedir + auxname)
endif
	
	if (!EqualWaves(WaveO, srcwave, 5))															// copy the modified wave only when something has changed
		Duplicate/O WaveO $(savedir + modname);	Wave srcwave = $(savedir + modname)				// for display purposes
		srcname = modname																		// for legend update
	endif
	WaveStats/Q WaveD
	if (V_numNans != numpnts(WaveD) && WaveExists(refwave))										// only if there is more than NaNs, copy the difference
		Duplicate/O WaveD $(savedir + difname);	Wave difwave = $(savedir + difname)
		if (saveAux && auxMode)
		Duplicate/O WaveA $(savedir + auxname);	Wave auxwave = $(savedir + auxname)
		endif
		doGraph = 1
	endif
	
	// ++++++++++++++++++++++++++ create and update summary graphs +++++++++++++++++++++++++++++++
	ControlInfo/W=$pPath FinishGraph
	if (V_Value && doGraph)
		Display/K=1/W=(200,50,700,300) as "Result of " + srcname_copy
		if (is2D)
			AppendImage difwave	
			ModifyImage $difname ctab= {*,*,Terrain,0}
			Legend/C/N=GraphInfoBox/F=0/A=LT/X=2.00/Y=2.00 
		else
			Legend/C/N=GraphInfoBox/F=0/A=LT/X=2.00/Y=2.00
				AppendToGraph srcwave
				AppendToGraph/C=(0,0,65280) refwave
				AppendToGraph/C=(0,0,0) difwave
			if (saveAux && auxMode)
				AppendToGraph/C=(0,40000,20000) auxwave
			endif
		endif
		string formula = ReplaceString("M",ReplaceString("R",StringByKey(kFormulaKey, note(WaveO), "=", "\r"),"ref"),"mod")
		if (strlen(formula))
			AppendText/N=GraphInfoBox "Equation: "+ReplaceString("_",StringFromList(setMode,kOutEnd),"")+" = "+formula
		endif
		ModifyGraph mirror(left)=1, mirror(bottom)=2, minor=1, standoff=0, ZisZ=1, notation(left)=1	
		Legend/C/N=ModificationInfoBox/F=0/A=RT/X=2.00/Y=2.00/J/B=1
	endif
	
	string graphList = WinList("!DiffGenMainGraph*", ";","WIN:1")
	for (i = 0; i < ItemsInList(graphList); i++)												// update the modification labels of all graphs with this data set
		string currWin = StringFromList(i, graphList)
		int matchWave = is2D ? Stringmatch(ImageNameList(currWin,";"),"*"+difname+"*") : Stringmatch(TraceNameList(currWin,";",1),"*"+srcname+"*")
		int matchAnno = Stringmatch(AnnotationList(currWin),"*ModificationInfoBox*")
		if (matchWave && matchAnno)
			UpdateModLegend(stg, currWin)
		endif
	endfor
	return 0
end